"""
FUSE operations for MoaT-micro-FS
"""

from __future__ import annotations

import anyio
import errno
import logging
import os
import stat
from contextlib import asynccontextmanager, suppress
from pathlib import PurePath, PurePosixPath

import pyfuse3
from pyfuse3 import (  # pylint: disable=E0611
    RENAME_EXCHANGE,
    RENAME_NOREPLACE,
    EntryAttributes,
    FileInfo,
    FUSEError,
)

from collections import defaultdict

# Typing

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from pyfuse3 import RequestContext, SetattrFields, StatvfsData

    from moat.lib.rpc import SubMsgSender

    from collections.abc import AsyncIterator, Callable, Sequence
    from typing import Any, NoReturn


logger = logging.getLogger(__name__)


class Operations(pyfuse3.Operations):  # pylint: disable=I1101
    "FUSE operations to delegate to a MicroPython client"

    # pylint: disable=too-many-public-methods
    supports_dot_lookup: bool = False
    enable_writeback_cache: bool = False
    enable_acl: bool = False

    max_read: int = 256
    max_write: int = 256

    _inode_path_map: dict[pyfuse3.InodeT, PurePath] = {pyfuse3.ROOT_INODE: PurePosixPath("/")}
    _path_inode_map: dict[PurePath, pyfuse3.InodeT] = {PurePosixPath("/"): pyfuse3.ROOT_INODE}
    _lookup_cnt: dict[pyfuse3.InodeT, Callable]
    _fd_inode_map: dict[int, pyfuse3.InodeT]
    _inode_fd_map: dict[pyfuse3.InodeT, int]
    _fd_open_count: dict[int, int]

    _dir_content: dict[int, tuple[pyfuse3.InodeT, Any, RequestContext]]
    _last_dir_fh: int = 0

    _link: SubMsgSender
    _last_inode: pyfuse3.InodeT = pyfuse3.ROOT_INODE

    def __init__(self, link: SubMsgSender):
        # pylint: disable=I1101
        super().__init__()
        self._inode_path_map = {pyfuse3.ROOT_INODE: PurePosixPath("/")}
        self._path_inode_map = {PurePosixPath("/"): pyfuse3.ROOT_INODE}
        self._lookup_cnt = defaultdict(lambda: 0)
        self._fd_inode_map = dict()
        self._inode_fd_map = dict()
        self._fd_open_count = dict()
        self._dir_content = dict()

        self._link = link

    def f_open(self, fd, inode):
        "remember opened file"
        self._fd_inode_map[fd] = inode
        self._inode_fd_map[inode] = fd

    def f_close(self, fd):
        "close opened file"
        i = self._fd_inode_map.pop(fd)
        del self._inode_fd_map[i]

    def i_path(self, inode: pyfuse3.InodeT) -> PurePath:
        "return path for inode"
        try:
            return self._inode_path_map[inode]
        except KeyError:
            logger.debug("NotFound: i=%r", inode)
            raise FUSEError(errno.ENOENT) from None

    def i_add(self, path: PurePath, inode: pyfuse3.InodeT | None = None) -> pyfuse3.InodeT:
        "invent new inode number"
        try:
            return self._path_inode_map[path]
        except KeyError:
            if inode is None:
                self._last_inode += 1
                inode = pyfuse3.InodeT(self._last_inode)
            self._inode_path_map[inode] = path
            self._path_inode_map[path] = inode
            return inode

    def i_del(self, inode: pyfuse3.InodeT) -> None:
        "drop inode number"
        try:
            p = self._inode_path_map.pop(inode)
        except KeyError:
            pass
        else:
            del self._path_inode_map[p]

    def raise_error(self, err: BaseException, inode: pyfuse3.InodeT | None = None) -> NoReturn:
        "translate generic exception to FUSE exception"
        logger.warning("Error: %r", err)
        if isinstance(err, FileNotFoundError):
            if inode is not None:
                self.i_del(inode)
            raise FUSEError(errno.ENOENT)
        if isinstance(err, FileExistsError):
            raise FUSEError(errno.EEXIST)
        raise FUSEError(errno.EIO)

    @staticmethod
    def _name(name: pyfuse3.FileNameT) -> str:
        if isinstance(name, (bytes, bytearray, memoryview)):
            return bytes(name).decode("utf-8")
        return str(name)

    async def lookup(
        self, parent_inode: pyfuse3.InodeT, name: pyfuse3.FileNameT, ctx: RequestContext
    ) -> EntryAttributes:
        """Look up a directory entry by name and get its attributes.

        This method should return an `~pyfuse3.EntryAttributes` instance for the
        directory entry *name* in the directory with inode *parent_inode*.

        If there is no such entry, the method should either return an
        `~pyfuse3.EntryAttributes` instance with zero ``st_ino`` value (in which case
        the negative lookup will be cached as specified by ``entry_timeout``),
        or it should raise `~pyfuse3.FUSEError` with an errno of `errno.ENOENT` (in this
        case the negative result will not be cached).

        The file system must be able to handle lookups for :file:`.` and
        :file:`..`, no matter if these entries are returned by `readdir` or not.

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """
        p = self.i_path(parent_inode) / self._name(name)
        try:
            return await self.getattr(self._path_inode_map[p], ctx)
        except KeyError:
            try:
                d = await self._link.stat(str(p))
            except Exception as err:  # pylint: disable=broad-exception-caught
                self.raise_error(err)
            else:
                inode = self.i_add(p)
                return await self._getattr(inode, ctx, _res=d)

    async def forget(self, inode_list: Sequence[tuple[pyfuse3.InodeT, int]]) -> None:
        """Decrease lookup counts for inodes in *inode_list*

        *inode_list* is a list of ``(inode, nlookup)`` tuples. This method
        should reduce the lookup count for each *inode* by *nlookup*.

        If the lookup count reaches zero, the inode is currently not known to
        the kernel. In this case, the file system will typically check if there
        are still directory entries referring to this inode and, if not, remove
        the inode.

        If the file system is unmounted, it may not have received `forget` calls
        to bring all lookup counts to zero. The filesystem needs to take care to
        clean up inodes that at that point still have non-zero lookup count
        (e.g. by explicitly calling :meth:`forget` with the current lookup
        count for every such inode after :func:`pyfuse3.main` has returned).

        This method must not raise any exceptions (not even `~pyfuse3.FUSEError`), since
        it is not handling a particular client request.
        """

        for i, _n in inode_list:
            with suppress(KeyError):
                self.i_del(i)

    async def _getattr(
        self, inode: pyfuse3.InodeT, _ctx: RequestContext, _res=None
    ) -> EntryAttributes:
        """Get attributes for *inode*

        This method should return an `~pyfuse3.EntryAttributes` instance with the
        attributes of *inode*. The `~pyfuse3.EntryAttributes.entry_timeout` attribute is
        ignored in this context.
        """

        p = self.i_path(inode)
        if _res is None:
            try:
                d = await self._link.stat(str(p))
            except Exception as err:  # pylint: disable=broad-exception-caught
                self.raise_error(err, inode)
        else:
            d = _res

        r = EntryAttributes()
        t = d["t"]
        r.st_ino = pyfuse3.InodeT(inode)
        r.entry_timeout = 300
        r.attr_timeout = 300

        r.st_size = 0
        if d["m"] == "d":
            r.st_mode = pyfuse3.ModeT(stat.S_IFDIR | 0o777)
        elif d["m"] == "f":
            r.st_mode = pyfuse3.ModeT(stat.S_IFREG | 0o666)
            r.st_size = d["s"]
        r.st_mtime_ns = t * 1_000_000_000
        r.st_ctime_ns = t * 1_000_000_000
        r.st_atime_ns = t * 1_000_000_000
        r.st_birthtime_ns = t * 1_000_000_000
        r.st_size = d.get("s", 0)
        r.st_blksize = 256
        r.st_blocks = (r.st_size - 1) // r.st_blksize + 1
        return r

    async def getattr(self, inode: pyfuse3.InodeT, ctx: RequestContext) -> EntryAttributes:
        """Get attributes for *inode*."""
        return await self._getattr(inode, ctx)

    async def setattr(
        self,
        inode: pyfuse3.InodeT,
        attr: EntryAttributes,
        fields: SetattrFields,
        fh,
        ctx: RequestContext,
    ) -> EntryAttributes:
        """Change attributes of *inode*

        *fields* specifies which attributes are to be updated. *attr*
        contains the new values of changed attributes for *inode*, and
        undefined values for all other attributes.

        Most file systems will additionally set the
        `~pyfuse3.EntryAttributes.st_ctime_ns` attribute to the current time (to
        indicate that the inode metadata was changed).

        If the syscall that is being processed received a file descriptor
        argument (like e.g. :manpage:`ftruncate(2)` or :manpage:`fchmod(2)`),
        *fh* will be the file handle returned by the corresponding call to the
        `open` handler. If the syscall was path based (like
        e.g. :manpage:`truncate(2)` or :manpage:`chmod(2)`), *fh* will be
        `None`.

        The method should return an `~pyfuse3.EntryAttributes` instance (containing both
        the changed and unchanged values).
        """

        logger.warning(
            "NotImpl: setattr: i=%r a=%r f=%r h=%r ctx=%r",
            inode,
            attr,
            fields,
            fh,
            ctx,
        )
        raise FUSEError(errno.ENOSYS)

    async def readlink(self, inode: int, ctx: RequestContext):
        """Return target of symbolic link *inode*."""

        logger.warning("NotImpl: readlink: i=%r ctx=%r", inode, ctx)
        raise FUSEError(errno.ENOSYS)

    async def mknod(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        mode: pyfuse3.ModeT,
        rdev: int,
        ctx: RequestContext,
    ) -> EntryAttributes:
        """Create (possibly special) file

        This method must create a (special or regular) file *name* in the
        directory with inode *parent_inode*. Whether the file is special or
        regular is determined by its *mode*. If the file is neither a block nor
        character device, *rdev* can be ignored.

        The method must return an `~pyfuse3.EntryAttributes` instance with the attributes
        of the newly created directory entry.

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """

        logger.warning(
            "NotImpl: mknod: p=%r n=%r m=%r d=%r ctx=%r",
            parent_inode,
            name,
            mode,
            rdev,
            ctx,
        )
        raise FUSEError(errno.ENOSYS)

    async def mkdir(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        mode: pyfuse3.ModeT,
        ctx: RequestContext,
    ) -> EntryAttributes:
        """Create a directory

        This method must create a new directory *name* with mode *mode* in the
        directory with inode *parent_inode*.

        This method must return an `~pyfuse3.EntryAttributes` instance with the
        attributes of the newly created directory entry.

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """
        mode, ctx  # noqa:B018
        p = self.i_path(parent_inode) / self._name(name)
        try:
            await self._link.mkdir(str(p))
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.raise_error(err)
        return await self._getattr(self.i_add(p), ctx)

    async def unlink(
        self, parent_inode: pyfuse3.InodeT, name: pyfuse3.FileNameT, ctx: RequestContext
    ) -> None:
        """Remove a (possibly special) file

        This method must remove the (special or regular) file *name* from the
        direcory with inode *parent_inode*.

        If the inode associated with *file* (i.e., not the *parent_inode*) has a
        non-zero lookup count, or if there are still other directory entries
        referring to this inode (due to hardlinks), the file system must remove
        only the directory entry (so that future calls to `readdir` for
        *parent_inode* will no longer include *name*, but e.g. calls to
        `getattr` for *file*'s inode still succeed). (Potential) removal of the
        associated inode with the file contents and metadata must be deferred to
        the `forget` method to be carried out when the lookup count reaches zero
        (and of course only if at that point there are no more directory entries
        associated with the inode either).

        """

        ctx  # noqa:B018
        p = self.i_path(parent_inode) / self._name(name)
        try:
            await self._link.rm(str(p))
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.raise_error(err)

    async def rmdir(
        self, parent_inode: pyfuse3.InodeT, name: pyfuse3.FileNameT, ctx: RequestContext
    ) -> None:
        """Remove directory *name*

        This method must remove the directory *name* from the direcory with
        inode *parent_inode*. If there are still entries in the directory,
        the method should raise ``FUSEError(errno.ENOTEMPTY)``.

        If the inode associated with *name* (i.e., not the *parent_inode*) has a
        non-zero lookup count, the file system must remove only the directory
        entry (so that future calls to `readdir` for *parent_inode* will no
        longer include *name*, but e.g. calls to `getattr` for *file*'s inode
        still succeed). Removal of the associated inode holding the directory
        contents and metadata must be deferred to the `forget` method to be
        carried out when the lookup count reaches zero.

        (Since hard links to directories are not allowed by POSIX, this method
        is not required to check if there are still other directory entries
        refering to the same inode. This conveniently avoids the ambigiouties
        associated with the ``.`` and ``..`` entries).
        """

        ctx  # noqa:B018
        p = self.i_path(parent_inode) / self._name(name)
        try:
            await self._link.rmdir(str(p))
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.raise_error(err)

    async def symlink(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        target: pyfuse3.FileNameT,
        ctx: RequestContext,
    ) -> EntryAttributes:
        """Create a symbolic link

        This method must create a symbolink link named *name* in the directory
        with inode *parent_inode*, pointing to *target*.

        The method must return an `~pyfuse3.EntryAttributes` instance with the attributes
        of the newly created directory entry.

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """

        logger.warning("NotImpl: symlink: p=%r n=%r t=%r ctx=%r", parent_inode, name, target, ctx)
        raise FUSEError(errno.ENOSYS)

    async def rename(
        self,
        parent_inode_old: pyfuse3.InodeT,
        name_old: str,
        parent_inode_new: pyfuse3.InodeT,
        name_new: str,
        flags: pyfuse3.FlagT,
        ctx: RequestContext,
    ) -> None:
        """Rename a directory entry.

        This method must rename *name_old* in the directory with inode
        *parent_inode_old* to *name_new* in the directory with inode
        *parent_inode_new*.  If *name_new* already exists, it should be
        overwritten.

        *flags* may be `~pyfuse3.RENAME_EXCHANGE` or `~pyfuse3.RENAME_NOREPLACE`. If
        `~pyfuse3.RENAME_NOREPLACE` is specified, the filesystem must not overwrite
        *name_new* if it exists and return an error instead. If
        `~pyfuse3.RENAME_EXCHANGE` is specified, the filesystem must atomically exchange
        the two files, i.e. both must exist and neither may be deleted.

        Let the inode associated with *name_old* in *parent_inode_old* be
        *inode_moved*, and the inode associated with *name_new* in
        *parent_inode_new* (if it exists) be called *inode_deref*.

        If *inode_deref* exists and has a non-zero lookup count, or if there are
        other directory entries referring to *inode_deref*), the file system
        must update only the directory entry for *name_new* to point to
        *inode_moved* instead of *inode_deref*.  (Potential) removal of
        *inode_deref* (containing the previous contents of *name_new*) must be
        deferred to the `forget` method to be carried out when the lookup count
        reaches zero (and of course only if at that point there are no more
        directory entries associated with *inode_deref* either).
        """

        try:
            p = self.i_path(parent_inode_old) / name_old
            q = self.i_path(parent_inode_new) / name_new

            if flags == 0:
                await self._link.mv(s=str(p), d=str(q))
            elif flags == RENAME_EXCHANGE:
                r = f"_xfn_{os.getpid()}"  # in the root
                await self._link.mv(s=str(p), d=str(q), x=str(r))
            elif flags == RENAME_NOREPLACE:
                await self._link.mv(s=str(p), d=str(q), n=True)
            else:
                logger.warning(
                    "NotImpl: rename: p=%r n=%r d=%r n=%r f=%r ctx=%r",
                    parent_inode_old,
                    name_old,
                    parent_inode_new,
                    name_new,
                    flags,
                    ctx,
                )
                raise FUSEError(errno.ENOSYS)
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.raise_error(err)

    async def link(
        self, inode, new_parent_inode, new_name, ctx: RequestContext
    ) -> EntryAttributes:
        """Create directory entry *name* in *parent_inode* refering to *inode*.

        The method must return an `~pyfuse3.EntryAttributes` instance with the
        attributes of the newly created directory entry.

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """

        logger.warning(
            "NotImpl: link: i=%r p=%r n=%r ctx=%r",
            inode,
            new_parent_inode,
            new_name,
            ctx,
        )
        raise FUSEError(errno.ENOSYS)

    async def open(
        self, inode: pyfuse3.InodeT, flags: pyfuse3.FlagT, ctx: RequestContext
    ) -> FileInfo:
        """Open a inode *inode* with *flags*.

        *flags* will be a bitwise or of the open flags described in the
        :manpage:`open(2)` manpage and defined in the `os` module (with the
        exception of ``O_CREAT``, ``O_EXCL``, ``O_NOCTTY`` and ``O_TRUNC``)

        This method must return a `~pyfuse3.FileInfo` instance. The
        `~pyfuse3.FileInfo.fh` field must contain an integer file handle,
        which will be passed to the :meth:`read`, :meth:`write`,
        :meth:`flush`, :meth:`fsync` and :meth:`release` methods to
        identify the open file. The `~pyfuse3.FileInfo` instance may also
        have relevant configuration attributes set; see the `~pyfuse3.FileInfo`
        documentation for more information.
        """
        ctx  # noqa:B018
        fh = FileInfo()
        if flags & os.O_RDWR:
            m = "a+" if flags & os.O_APPEND else "r+"
        elif flags & os.O_WRONLY:
            m = "a" if flags & os.O_APPEND else "w"
        else:
            m = "r"

        try:
            fd = await self._link.open(str(self.i_path(inode)), m=m)
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.raise_error(err)
        self.f_open(fd, inode)
        fh.fh = fd
        return fh

    async def read(self, fh, off, size):
        """Read *size* bytes from *fh* at position *off*

        *fh* will by an integer filehandle returned by a prior `open` or
        `create` call.

        This function should return exactly the number of bytes requested except
        on EOF or error, otherwise the rest of the data will be substituted with
        zeroes.
        """

        if size <= self.max_read:
            return await self._link.rd(fh, off, n=size)

        # OWCH. Need to break that large read up.

        data = []
        while size > 0:
            dl = min(size, self.max_read)
            buf = await self._link.rd(f=fh, o=off, n=dl)
            if buf == b"":
                break
            data.append(buf)
            if len(buf) < dl:
                break
            size -= dl
            off += dl
        return b"".join(data)

    async def write(self, fh, off, buf):
        """Write *buf* into *fh* at *off*

        *fh* will by an integer filehandle returned by a prior `open` or
        `create` call.

        This method must return the number of bytes written. However, unless the
        file system has been mounted with the ``direct_io`` option, the file
        system *must* always write *all* the provided data (i.e., return
        ``len(buf)``).
        """

        if len(buf) <= self.max_write:
            return await self._link.wr(fh, off, d=buf)

        # OWCH. Break that up.
        sent = 0
        while sent < len(buf):
            sn = await self._link.wr(
                f=fh,
                d=buf[sent : sent + self.max_write],
                o=off + sent,
            )
            sent += sn
            if sn < self.max_write:
                break
        return sent

    async def flush(self, fh):
        """Handle close() syscall.

        *fh* will by an integer filehandle returned by a prior `open` or
        `create` call.

        This method is called whenever a file descriptor is closed. It may be
        called multiple times for the same open file (e.g. if the file handle
        has been duplicated).
        """

        # pylint: disable=unnecessary-pass

    async def release(self, fh):
        """Release open file

        This method will be called when the last file descriptor of *fh* has
        been closed, i.e. when the file is no longer opened by any client
        process.

        *fh* will by an integer filehandle returned by a prior `open` or
        `create` call. Once `release` has been called, no future requests for
        *fh* will be received (until the value is re-used in the return value of
        another `open` or `create` call).

        This method may return an error by raising `~pyfuse3.FUSEError`, but the error
        will be discarded because there is no corresponding client request.
        """
        self.f_close(fh)
        await self._link.cl(fh)

    async def fsync(self, fh, datasync):
        """Flush buffers for open file *fh*

        If *datasync* is true, only the file contents should be
        flushed (in contrast to the metadata about the file).

        *fh* will by an integer filehandle returned by a prior `open` or
        `create` call.
        """
        # pylint: disable=unnecessary-pass

    async def opendir(self, inode, ctx: RequestContext):
        """Open the directory with inode *inode*

        This method should return an integer file handle. The file handle will
        be passed to the `readdir`, `fsyncdir` and `releasedir` methods to
        identify the directory.
        """
        # TODO: send an extended "ls" command and cache the state and timestamp

        p = self.i_path(inode)
        try:
            dc = await self._link.ls(str(p))
        except Exception as err:  # pylint:disable=broad-exception-caught
            self.raise_error(err)

        self._last_dir_fh += 1
        fh = self._last_dir_fh
        self._dir_content[fh] = (inode, dc, ctx)
        return fh

    async def readdir(self, fh, start_id, token):
        """Read entries in open directory *fh*.

        This method should list the contents of directory *fh* (as returned by a
        prior `opendir` call), starting at the entry identified by *start_id*.

        Instead of returning the directory entries directly, the method must
        call :meth:`~pyfuse3.readdir_reply` for each directory entry. If
        :meth:`~pyfuse3.readdir_reply` returns True, the file system must
        increase the lookup count for the provided directory entry by one
        and call :meth:`~pyfuse3.readdir_reply` again for the next entry (if
        any). If :meth:`~pyfuse3.readdir_reply` returns False, the lookup count
        must *not* be increased and the method should return without
        further calls to :meth:`~pyfuse3.readdir_reply`.

        The *start_id* parameter will be either zero (in which case listing
        should begin with the first entry) or it will correspond to a value that
        was previously passed by the file system to the :meth:`~pyfuse3.readdir_reply`
        function in the *next_id* parameter.

        If entries are added or removed during a `readdir` cycle, they may or
        may not be returned. However, they must not cause other entries to be
        skipped or returned more than once.

        :file:`.` and :file:`..` entries may be included but are not
        required. However, if they are reported the filesystem *must not*
        increase the lookup count for the corresponding inodes (even if
        :meth:`~pyfuse3.readdir_reply` returns True).
        """

        dir_inode, dc, ctx = self._dir_content[fh]
        p = self.i_path(dir_inode)

        for name in dc[start_id:]:
            inode = self.i_add(p / name)
            try:
                attr = await self.getattr(inode, ctx)
            except FileNotFoundError:
                self.i_del(inode)

            start_id += 1
            if not pyfuse3.readdir_reply(  # pylint: disable=I1101
                token,
                name.encode("utf-8"),
                attr,
                start_id,
            ):
                break

    async def releasedir(self, fh):
        """Release open directory

        This method will be called exactly once for each `opendir` call. After
        *fh* has been released, no further `readdir` requests will be received
        for it (until it is opened again with `opendir`).
        """

        del self._dir_content[fh]

    async def fsyncdir(self, fh, datasync):
        """Flush buffers for open directory *fh*

        If *datasync* is true, only the directory contents should be
        flushed (in contrast to metadata about the directory itself).
        """
        # pylint: disable=unnecessary-pass

    async def statfs(self, ctx: RequestContext) -> StatvfsData:
        """Get file system statistics

        The method must return an appropriately filled `~pyfuse3.StatvfsData` instance.
        """

        logger.warning("NotImpl: statfs: ctx=%r", ctx)
        raise FUSEError(errno.ENOSYS)

    def stacktrace(self):
        """Asynchronous debugging

        This method will be called when the ``fuse_stacktrace`` extended
        attribute is set on the mountpoint. The default implementation logs the
        current stack trace of every running Python thread. This can be quite
        useful to debug file system deadlocks.
        """

        # pylint: disable=import-outside-toplevel
        import sys  # noqa: PLC0415
        import traceback  # noqa: PLC0415

        code = []
        for threadId, frame in sys._current_frames().items():  # noqa:SLF001 pylint:disable=protected-access
            code.append(f"\n# ThreadID: {threadId}")
            for filename, lineno, name, line in traceback.extract_stack(frame):
                code.append(f"{os.path.basename(filename)}:{lineno}, in {name}")
                if line:
                    code.append("    {line.strip()}")

        logger.error("\n".join(code))

    async def setxattr(self, inode, name, value, ctx: RequestContext):
        """Set extended attribute *name* of *inode* to *value*.

        The attribute may or may not exist already. Both *name* and *value* will
        be of type `bytes`. *name* is guaranteed not to contain zero-bytes
        (``\\0``).
        """

        logger.warning("NotImpl: setxattr: i=%r n=%r v=%r ctx=%r", inode, name, value, ctx)
        raise FUSEError(errno.ENOSYS)

    async def getxattr(self, inode: int, name: bytes, ctx: RequestContext) -> bytes:
        """Return extended attribute *name* of *inode*

        If the attribute does not exist, the method must raise `~pyfuse3.FUSEError`
        with an error code of `~errno.ENOATTR`. *name* will be of type `bytes`,
        but is guaranteed not to contain zero-bytes (``\\0``).
        """

        logger.warning("NotImpl: getxattr: i=%r n=%r ctx=%r", inode, name, ctx)
        raise FUSEError(errno.ENOSYS)

    async def listxattr(
        self, inode: pyfuse3.InodeT, ctx: RequestContext
    ) -> Sequence[pyfuse3.XAttrNameT]:
        """Get list of extended attributes for *inode*

        This method must return a sequence of `bytes` objects.  The objects must
        not include zero-bytes (``\\0``).
        """

        logger.warning("NotImpl: listxattr: i=%r ctx=%r", inode, ctx)
        raise FUSEError(errno.ENOSYS)

    async def removexattr(self, inode: int, name: bytes, ctx: RequestContext) -> None:
        """Remove extended attribute *name* of *inode*

        If the attribute does not exist, the method must raise `~pyfuse3.FUSEError` with
        an error code of `~errno.ENOATTR`. *name* will be of type `bytes`, but is
        guaranteed not to contain zero-bytes (``\\0``).
        """

        logger.warning("NotImpl: removexattr: i=%r n=%r ctx=%r", inode, name, ctx)
        raise FUSEError(errno.ENOSYS)

    async def access(self, inode: int, mode: int, ctx: RequestContext) -> bool:
        """Check if requesting process has *mode* rights on *inode*.

        If the ``default_permissions`` mount option is given, this method is not
        called.

        When implementing this method, the :func:`~pyfuse3.get_sup_groups`
        function may be useful.
        """

        logger.warning("NotImpl: access: i=%r m=%r ctx=%r", inode, mode, ctx)
        raise FUSEError(errno.ENOSYS)

    async def create(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        mode: pyfuse3.ModeT,
        flags: pyfuse3.FlagT,
        ctx: RequestContext,
    ) -> tuple[FileInfo, EntryAttributes]:
        """Create a file with permissions *mode* and open it with *flags*

        (Successful) execution of this handler increases the lookup count for
        the returned inode by one.
        """

        mode  # noqa:B018
        p = self.i_path(parent_inode) / self._name(name)
        try:
            i = self._path_inode_map[p]
            gen = False
        except KeyError:
            i = self.i_add(p)
            gen = True

        try:
            r = await self._link.new(str(p))
        except Exception as err:  # pylint: disable=broad-exception-caught
            if gen:
                self.i_del(i)
            self.raise_error(err)
        else:
            i = self.i_add(p)
            h = await self.open(i, flags, ctx)
            a = await self._getattr(i, ctx, _res=r)
            return (h, a)


@asynccontextmanager
async def wrap(
    link: SubMsgSender, path: str | PurePath | anyio.Path, blocksize: int = 0, debug: int = 1
) -> AsyncIterator[None]:
    """
    Context manager that mounts a satellite file system locally
    """
    operations = Operations(link)
    if blocksize:
        operations.max_read = blocksize
        operations.max_write = blocksize

    logger.debug("Mounting...")
    # Whether max_read needs to be set is anybody's guess, so try both
    for mx in (False, True):
        fuse_options = set(pyfuse3.default_options)  # pylint: disable=I1101
        fuse_options.add("fsname=moat_fs")
        fuse_options.add(f"max_read={operations.max_read if mx else 0}")
        # fuse_options.add(f'max_write={operations.max_write}')
        if debug > 1:
            fuse_options.add("debug")
        try:
            pyfuse3.init(operations, str(path), fuse_options)  # pylint: disable=I1101
        except Exception:
            if mx:
                raise
        else:
            break

    logger.debug("Entering main loop..")
    async with anyio.create_task_group() as tg:

        async def fuse_main():
            try:
                await pyfuse3.main()
            except anyio.get_cancelled_exc_class():
                pyfuse3.close(unmount=True)
                raise
            except BaseExceptionGroup as exc:
                exc = exc.split(anyio.get_cancelled_exc_class())[1]
                pyfuse3.close(unmount=exc is None)
                raise
            except BaseException:
                pyfuse3.close(unmount=False)
                raise
            else:
                pyfuse3.close(unmount=True)

        try:
            tg.start_soon(fuse_main)  # pylint: disable=I1101
            yield None

        finally:
            tg.cancel_scope.cancel()



"""
TMP102 i²c sensor
"""

from __future__ import annotations

import struct

from moat.lib.micro import sleep_ms
from moat.micro.util import Sensor

__all__ = ["Cmd"]


class Cmd(Sensor):
    """
    This is the front-end for a TMP102 i²c temperature sensor.

    Config:
        bus: Path to the bus to use
        adr: Bus address to access
    """

    doc = dict(Sensor.doc)
    doc_c = doc.get("_c", {})
    if not isinstance(doc_c, dict):
        doc_c = {}
    else:
        doc_c = dict(doc_c)
    doc_c["bus"] = "path:i2c bus"
    doc_c["adr"] = "path:bus address"
    doc["_c"] = doc_c

    async def task(self):
        "wrapper, for i2c bus access"
        self.adr = self.cfg["adr"]
        async with self.root.sub_at(self.cfg["bus"]) as self.bus:
            await super().task()

    async def read(self):
        "read sensor"
        await self.bus.wr(self.adr, bytes((1, 0x81, 0x10)))
        # OneShot+Shutdown, Extended
        for _ in range(20):
            await sleep_ms(3)
            res = await self.bus.wrrd(self.adr, bytes((1,)), 2)
            if res[0] & 0x80:
                break
        else:
            raise RuntimeError("No conv")
        res = await self.bus.wrrd(self.adr, bytes((0,)), 2)

        (t,) = struct.unpack(">h", res)
        if not (t & 1):
            raise ValueError(t)
        t -= 1
        return t / 128



"""
Test runner
"""

from __future__ import annotations

import os
from contextlib import suppress
from pathlib import Path

import moat.micro
from moat.util import attrdict
from moat.lib.codec import get_codec
from moat.lib.rpc._test import temp_dir

# from moat.micro.main import Request, get_link, get_link_serial
# from moat.micro.proto.multiplex import Multiplexer
from moat.lib.stream import ProcessBuf

required = [
    "__future__",
    "copy",
    "errno",
    "pprint",
    "typing",
    "types",
    "functools",
    "contextlib",
    "ucontextlib",
    "collections",
    "inspect",
]


def rlink(s, d):
    "recursive linking"
    if s.is_file():
        with suppress(FileExistsError):
            d.symlink_to(s)
    else:
        with suppress(FileExistsError):
            d.mkdir()
        for f in s.iterdir():
            rlink(s / f.name, d / f.name)


class MpyBuf(ProcessBuf):
    """
    A stream that links to MicroPython.

    Parameters:
        dupterm(bool): Use a MPy variant that supports dupterm.
                       The downside is that such a variant currently
                       does not support writing to stderr …
    """

    async def setup(self):
        codec = get_codec("std-cbor")
        dupterm = self.cfg.get("dupterm", False)
        pre = Path(__file__).parents[2]
        upy = pre / "ext/micropython"

        root = self.cfg.get("cwd", None)
        if root is None:
            root = temp_dir.get() / "root"
        else:
            root = Path(root).absolute()  # noqa: ASYNC240, RUF100 Test only
        lib = root / "stdlib"
        lib2 = root / "lib"
        with suppress(FileExistsError):
            root.mkdir()
        with suppress(FileExistsError):
            lib.mkdir()
        with suppress(FileExistsError):
            lib2.mkdir()
        with suppress(FileExistsError):
            (root / "tests").symlink_to(Path("tests").absolute())  # noqa: ASYNC240, RUF100 Test only

        std = (upy / "lib/micropython-lib/python-stdlib").absolute()
        ustd = (upy / "lib/micropython-lib/micropython").absolute()
        for req in required:
            if (std / req).exists():
                rlink(std / req, lib)
            elif (ustd / req).exists():
                rlink(ustd / req, lib)
            else:
                raise FileNotFoundError(std / req)

        aio = Path("lib/micropython/extmod/asyncio").absolute()  # noqa: ASYNC240, RUF100 Test only
        with suppress(FileExistsError):
            (lib / "asyncio").symlink_to(aio)

        libp = [lib, lib2]
        for p in moat.micro.__path__:
            p = Path(p) / "_embed"  # noqa:PLW2901
            if p.exists():
                libp.append(p)
            if (p / "lib").exists():
                libp.append(p / "lib")
        libp.append(".frozen")

        self.env = {
            "MICROPYPATH": os.pathsep.join(str(x) for x in libp),
        }
        self.cwd = root

        with (root / "moat.cfg").open("wb") as f:
            f.write(codec.encode(self.cfg["cfg"]))
        with (root / "moat.state").open("w") as f:
            f.write(self.cfg.get("state", "once"))
        if self.cfg.get("large", True):
            with (root / "moat.lrg").open("wb") as f:
                pass

        rlink(libp[2] / "boot.py", root / "boot.py")
        rlink(libp[2] / "main_unix.py", root / "main.py")
        self.argv = [
            # "strace", "-s300", "-o/tmp/mpy.log",
            pre / "build" / ("mpy-unix" + ("-dup" if dupterm else "")) / "micropython",
            "-e",
        ]

        await super().setup()


# Fake "machine" module

machine = attrdict()


class FakeI2C:
    def __init__(self, c, d, **_):
        self._c = c
        self._d = d


class FakePin:
    def __init__(self, pin, **_):
        self._pin = pin


machine.Pin = FakePin
machine.I2C = FakeI2C
machine.SoftI2C = FakeI2C



"""
Main interface to the real-time clock
"""

from __future__ import annotations

from moat.util import NotGiven, de_async, merge
from moat.lib.micro import log

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import builtins

    from moat.lib.rpc import RootCmd

    from collections.abc import Awaitable
    from typing import Any, ClassVar


class NotSync(RuntimeError):
    """
    This RTC implementation can't access its data synchronously.
    """


class _RTC:
    """
    Singleton class for RTC access

    Config:
        back: list of backend names to use

    """

    backends: list[RTCBase]
    root = None

    def __repr__(self):
        return "<RTC>"

    def init(self, cfg: dict | None = None):
        self.cfg = cfg or {"use": ["mem", "fs"]}
        self._be = {}
        self._setup()

        try:
            cfg = de_async(self.get, "rtc", sync=True)
        except KeyError:
            pass
        except Exception as exc:
            log("No RTC config: %r", exc)
        else:
            merge(self.cfg, cfg)
            self._setup()

    def set_root(self, root: RootCmd):
        self.root = root
        self._setup()

    def _setup(self, err: bool = False):
        self.backends = bes = []
        nbe = set(self._be.keys())
        mp = "moat.micro.rtc."  # module name prefix

        for i, c in enumerate(self.cfg.get("use", [])):
            if isinstance(c, str):
                c = {"name": c}  # noqa:PLW2901
            n = c.get("name", str(i))
            if (be := self._be.get(n, None)) is not None:
                be.reload()
            else:
                try:
                    m = c.get("mod", n)
                    if "." not in m:
                        m = mp + m

                    cls = __import__(m, globals(), None, ("RTC",), 0).RTC
                    be = cls(c)  # RTCBase instance
                    if be.is_ASYNC is False:
                        be.is_ASYNC = None
                    elif be.is_ASYNC is None:
                        be.is_ASYNC = True
                        self._be[n] = be
                        bes.append(be)
                        nbe.discard(n)
                        be = cls(c)  # RTCBase instance
                        be.is_ASYNC = False
                        n += "_sync"
                        be.cfg["name"] = n

                except Exception as exc:
                    if err:
                        raise
                    if not isinstance(exc, ImportError) or m != mp + "mem":
                        log("RTC error for %s", n, err=exc)
                    continue
            self._be[n] = be
            bes.append(be)
            nbe.discard(n)
        for n in nbe:
            be = self._be.pop(n)

    def reload(self):
        self._setup(err=True)

    def state(self, name: str) -> State:
        """
        State accessor.
        """
        return State(self, name)

    def get_sync(self, name, fs=None, default=NotGiven):
        """
        Synchronous getter for RTC data.

        Args:
            name: The key to retrieve.
            fs: Access the file system?
            default: Default value if key not found.

        Returns:
            The stored value, or default if not found.

        Raises:
            KeyError: If key not found and no default provided.
        """
        try:
            return de_async(self.get, name, fs=fs, sync=True)
        except KeyError:
            if default is NotGiven:
                raise
            return default

    def set_sync(self, name: str, value: Any, **kw):
        """
        Synchronous setter for RTC data.

        Args:
            name: The key to store.
            value: The value to store.

        Other keyword args are forwarded to :meth:`set`.
        """
        de_async(self.set, name, value, sync=True, **kw)

    async def get(self, name: str, fs: bool | None = None, sync: bool = False) -> Any:
        """
        Retrieve an RTC datum.

        Args:
            name: The data to get.
            fs: Access the file system?
            sync: if set, this call must not yield.
        """
        exc = None
        exc2 = None

        for be in self.backends:
            if fs is (not be.is_FS):
                continue
            if be.is_ASYNC is sync:
                continue
            try:
                res = await be.get(name)
                return res
            except KeyError:
                pass
            except NotSync:
                exc2 = NotSync(name)
            except Exception as err:
                exc = err
        if exc is not None:
            raise exc from None
        if exc2 is not None:
            raise exc2 from None
        raise KeyError(name) from None

    async def keys(self, fs: bool | None = None, sync: bool = False) -> builtins.set[str]:
        """
        Get the list of available/existing RTC keys.
        """
        res = set()

        for be in self.backends:
            if fs is (not be.is_FS):
                continue
            if be.is_ASYNC is sync:
                continue
            res.update(await be.keys())

        return res

    async def set(
        self,
        name: str,
        data: Any,
        fs: bool | None = None,
        sync: bool = False,
        all: bool | None = True,  # noqa:A002
    ) -> None:
        """
        Set an RTC datum.

        Args:
            name: The data to get.
            data: The data to write.
            fs: Access the file system?
            sync: Flag whether yielding is blocked.
            all: Write to all backends?

            *all* can be
            - `True`: write to all backends.
            - `False`: only use the first available backend.
            - `None`: write to the first matching backend. Erase this key
              from other backends.
        """
        seen = False
        exc = None
        if fs is None and all is None:
            try:
                await self.set(name, data, fs=False)
            except NotImplementedError:
                pass
            await self.set(name, data, fs=True)
            return

        for be in self.backends:
            if be.is_ASYNC is sync:
                continue
            try:
                if fs is (not be.is_FS):
                    continue
                if not await be.set(name, data):
                    continue
                seen = True

                if all is None:
                    data = NotGiven
                elif not all:
                    return
            except KeyError:
                continue
            except Exception as err:
                exc = err
        if exc is not None:
            raise exc from None
        if not seen:
            raise NotImplementedError(name)


RTC = _RTC()


class State:
    """
    Accessor for a single state.
    """

    def __init__(self, rtc: RTCBase | _RTC, name: str, fs: bool | None = None):
        self.rtc = rtc
        self.name = name
        self.fs = fs

    def get(self, **kw) -> Awaitable[Any]:
        "Getter."
        if "fs" not in kw:
            kw["fs"] = self.fs
        return self.rtc.get(self.name, **kw)

    def set(self, data: Any, **kw) -> Awaitable[None]:
        "Setter."
        if "fs" not in kw:
            kw["fs"] = self.fs
        return self.rtc.set(self.name, data, **kw)


class RTCBase:
    """
    Base class for RTC backends.

    Class variables:
        is_FS: This class accesses a (Flash) file system, thus should not be
               write often (if at all).
        is_ASYNC: Flag whether the implementation yields, i.e. cannot
                  be called synchronously.
                  If `None`, the loader shall create both a
                  sync and an async version.
    """

    is_FS: ClassVar[bool] = False
    is_ASYNC: bool | None = None

    def __init__(self, cfg):
        self.cfg = cfg
        sync = cfg.get("sync", None)
        if self.is_ASYNC is None:
            self.is_ASYNC = sync
        elif sync is (not self.is_ASYNC):
            raise ValueError("Sync settings don't match")

    def __repr__(self):
        return f"<{self._name}>"

    @property
    def _name(self):
        try:
            return self.cfg["name"]
        except KeyError:
            try:
                return self.cfg["mod"]
            except KeyError:
                return self.__class__.__name__

    async def get(self, name: str) -> Any:
        """
        Getter.
        """
        raise NotImplementedError

    async def set(self, name: str, data: Any) -> bool:
        """
        Setter.

        Returns:
            True if it could write.
        """
        raise NotImplementedError

    def state(self, name: str) -> State:
        """
        State accessor for a specific RTC.
        """
        return State(self, name)

    async def keys(self) -> builtins.set[str]:
        """
        Get all available keys.
        """
        raise NotImplementedError


def at(*a, **kw):
    """
    Setter for debugging.

    Usage: call ``at("something", or_other=42)`` at various places in your
    code. After a crash the data from the last such call will be available
    by calling ``RTC.get_sync("debug")``.
    """
    try:
        if (RTC.get_sync("state") + ",").startswith("safe,"):
            return
    except KeyError:
        pass
    data = (a, kw) if a and kw else (a or kw)
    try:
        RTC.set_sync("debug", data, fs=False)
    except NotImplementedError:
        pass



"""
Command tree support for MoaT commands
"""

from __future__ import annotations

from moat.util import NotGiven, attrdict
from moat.lib.micro import (
    AC_use,
    L,
    Lock,
    TaskGroup,
)
from moat.lib.rpc import BaseCmd

# Typing

from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from moat.lib.micro import _TaskGroupProto
    from moat.lib.path import PathElem
    from moat.lib.rpc import BaseMsgHandler

    from collections.abc import Awaitable, Callable


class BaseSuperCmd(BaseCmd):
    """
    A handler that can have a nested app (or more than one).

    Sets up a taskgroup for the sub-app(s) to run in.
    """

    tg: object | None = None
    app_lock: Lock | None = None

    async def setup(self) -> None:
        "setup apps"
        await super().setup()
        self.app_lock = Lock()
        tg = await AC_use(self, cast("AbstractAsyncContextManager[object]", TaskGroup()))
        self.tg = tg
        await AC_use(self, cast("_TaskGroupProto", tg).cancel)

    async def start_app(self, app: BaseCmd) -> None:
        """
        Run (or reload) this app.
        """

        async def _run(app):
            try:
                await app.run()
            finally:
                app.p_task = None

        if self.app_lock is None:
            raise RuntimeError("No lock")
        tg = self.tg
        if tg is None:
            raise RuntimeError("No taskgroup")
        tg = cast("_TaskGroupProto", tg)
        async with self.app_lock:
            if app.p_task:
                return
            try:

                async def _run_app() -> None:
                    await _run(app)

                t = await tg.spawn(_run_app)
                if app.p_task is False:
                    # set by .stop()
                    t.cancel()
                    app.p_task = None
                    return

                app.p_task = t
                if L:
                    await app.wait_started()
            except BaseException:
                app.p_task = None
                raise


class BaseSubCmd(BaseSuperCmd):
    """
    A handler for a directory.

    Apps have a hierarchical structure. This class serves as the equivalent
    of a subdirectory.

    How to create new entries is not specified in this class.
    """

    sub: dict[str, BaseCmd]

    def __init__(self, cfg):
        super().__init__(cfg)
        self.sub = {}

    if L:

        async def wait_ready(self, wait=True):
            """Delay until this subtree is up,

            Returns True if all sub-apps are stopped.
            """
            res = await super().wait_ready(wait=wait)
            if res is None:
                return None
            for app in list(self.sub.values()):
                if (w := await app.wait_ready(wait=wait)) is None:
                    return None
                res &= w
            return res

    async def attach(self, name, app) -> None:
        """
        Attach a sub-handler to me.

        An existing handler with this name is stopped.
        """
        oa = self.sub.pop(name, None)
        if app is not None:
            self.sub[name] = app
            app.attached(self, name)
        if oa is not None:
            await oa.stop()

    def detach(self, name) -> Awaitable:
        """
        Detach and stop a command handler.
        """
        return self.attach(name, None)

    async def reload(self):
        "reload apps"
        root = self.root
        if root is None:
            raise RuntimeError("No root")
        root.cfg_reloaded(self.cfg)

        await super().reload()
        async with TaskGroup() as tg:
            for app in list(self.sub.values()):
                tg.start_soon(app.reload)

    def find_sub(self, scmd: PathElem) -> BaseMsgHandler | Callable[..., object] | None:
        """
        Resolve a subcommand.

        This version uses the ``sub`` mapping.
        """
        if isinstance(scmd, str) and (sub := self.sub.get(scmd, None)) is not None:
            return sub

        return super().find_sub(scmd)

    doc_dir_ = dict(
        _d="list cmd subdirectory",
        d="dict(str,name):sub-apps",
        c=["str:commands"],
        v="bool:show hidden",
    )

    async def cmd_dir_(self, v=True):
        "dir: add subdirs"
        res = await super().cmd_dir_(v=v)
        dd = {}
        for k, v in self.sub.items():
            if not isinstance(k, str) or v is (k[-1] == "_"):
                continue
            try:
                dd[k] = v.cfg["app"]
            except (AttributeError, KeyError):
                dd[k] = v.__class__.__name__
        if dd:
            res["d"] = dd
        return res


class DirCmd(BaseSubCmd):
    """
    A BaseSubCmd handler with apps started by local configuration.

    Not typically subclassed.
    """

    doc = dict(_c=dict(_d="subdirectory", _n="app:sub-apps"))

    SKIP_RDY = True

    def __init__(self, cfg):
        super().__init__(cfg)
        self._lock = Lock()

    async def reload(self):
        "called after the config has been updated"
        async with self._lock:
            await super().reload()
            await self._setup_apps()

    cmd_upd_ = reload

    async def task(self):
        "Setup apps"
        await self._setup_apps()
        await super().task()

    async def _setup_apps(self):
        from moat.lib.rpc import LoadCmd  # noqa: PLC0415

        # log("Setup %s: %s", self, self.path)
        gcfg = self.cfg
        root = self.root
        if root is None:
            raise RuntimeError("No root")
        root.cfg_reloaded(gcfg)

        apps = {}
        for k, v in gcfg.items():
            try:
                if not v.get("running", True):
                    continue
                nam = v.get("app", None)
                if not isinstance(nam, str):
                    raise ValueError(f"Bad Config: app-{k}")  # noqa:TRY004
                apps[k] = v
            except (AttributeError, KeyError):
                continue

        # Zeroth, kill apps that are no longer live
        for name in list(self.sub.keys()):
            if name not in apps:
                await self.detach(name)

        # First, setup the app data structures
        for name, v in apps.items():
            if name in self.sub:
                continue
            if v is NotGiven:
                continue

            cfg = gcfg.get(name, attrdict())
            await self.attach(name, LoadCmd(cfg))

        # Second, run them all.
        # For existing apps, tell it to update its configuration.
        for app in self.sub.values():
            async with TaskGroup() as tg:

                async def _start(app: BaseCmd = app) -> None:
                    await self.start_app(app)

                tg.start_soon(_start)

        # Third, wait for them to be up.
        if L:
            for app in self.sub.values():
                if app.cfg.get("wait", True):
                    await app.wait_ready()

        # Finally, mark done.
        if L:
            self.set_ready()



"""
A heap of compatibility code that adapts CPython and MicroPython
to something roughly equivalent.
"""

from __future__ import annotations

import os as _os
import sys
from inspect import iscoroutine
from time import ticks_add, ticks_diff, ticks_ms

from async_queue import Queue, QueueEmpty, QueueFull  # noqa:F401

import asyncio
from micropython import const

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from typing import Never

Iterable = object  # TODO
Mapping = dict
MutableSequence = list
MutableMapping = dict


def _l():
    import os  # noqa: PLC0415

    try:
        os.stat("moat.lrg")
    except OSError:
        return False
    else:
        return True


L = _l()
del _l

Event = asyncio.Event
Lock = asyncio.Lock
sleep = asyncio.sleep
sleep_ms = asyncio.sleep_ms
TimeoutError = asyncio.TimeoutError  # noqa:A001
ModuleNotFoundError = ImportError  # noqa:A001
_run = asyncio.run
_tg = asyncio.TaskGroup
CancelledError = asyncio.CancelledError
ObjSequence = (list, tuple)


def breakpoint():  # noqa:A001
    "does nothing on µPy"
    pass


ExceptionGroup = asyncio.ExceptionGroup  # noqa: A001
BaseExceptionGroup = asyncio.BaseExceptionGroup  # noqa: A001

DEBUG = const(False)


class EndOfStream(Exception):
    "as from anyio"


class BrokenResourceError(Exception):
    "as from anyio"


try:
    from machine import Pin
except ImportError:  # µPy on Linux
    Pin_IN = "IN"
    Pin_OUT = "OUT"
else:
    Pin_IN = Pin.IN
    Pin_OUT = Pin.OUT

WouldBlock = (QueueFull, QueueEmpty)


def byte2utf8(buf):
    "UTF-8 decoder that works with memoryview et al."
    if not hasattr(buf, "decode"):
        buf = bytes(buf)
    return buf.decode("utf-8")


def print_exc(exc, file=None):
    "forwards to sys.print_exception"
    if file is None:
        file = sys.stderr
    sys.print_exception(exc, file)


def log(s, *x, err=None, write: bool = True):
    """
    Basic logger.debug/error call (depends on @err).

    Args:
        s:
            String to print, followed by %-formatting arguments like logging.debug.
        err:
            None for debug, True for error, or an exception if you want a
            traceback.

    """
    if x:
        s = s % x

    if write and err is not None:
        if "moat.exc" not in _os.listdir():
            with open("moat.exc", "w") as f:
                print(s, file=f)
                print_exc(err, f)

    print(s, file=sys.stderr)
    if isinstance(err, BaseException):
        print_exc(err)


def log_exc(e, s, *a):
    "Log an exception"
    log(s + ": %r", *a, e, err=e)


async def idle():
    "sleep forever"
    while True:
        await sleep(60 * 60 * 12)  # half a day


def wait_for(timeout, p, *a, **k):
    """
    asyncio.wait_for() but with sane calling convention
    """
    return asyncio.wait_for(p(*a, **k), timeout)


def wait_for_ms(timeout, p, *a, **k):
    """
    asyncio.wait_for_ms() but with sane calling convention
    """
    return asyncio.wait_for_ms(p(*a, **k), int(timeout))


class _MsecIter:
    tt = None

    def __init__(self, t, p, a, k):
        self.t = t
        self.p, self.a, self.k = p, a, k

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.tt is None:
            self.tt = ticks_add(ticks_ms(), self.t)
        else:
            tn = ticks_ms()
            if (td := ticks_diff(self.tt, tn)) > 0:
                await sleep_ms(td)
                self.tt = ticks_add(self.tt, self.t)
            else:
                # owch, delay too long
                self.tt = ticks_add(tn, self.t)
        if self.p is None:
            return None
        return await self.p(*self.a, **self.k)


def every_ms(t, p=None, *a, **k):
    "call a function every @t milliseconds"
    return _MsecIter(t, p, a, k)


def every(t, *a, **k):
    "call a function every @t seconds"
    return every_ms(int(t * 1000), *a, **k)


async def retry_ms(n, t, p=None, *a, _exc=Exception, **k):
    "call a function every @t milliseconds until it succeeds"
    while True:
        try:
            return await p(*a, **k)
        except _exc:
            if n >= 0:
                n -= 1
            if n == 0:
                raise
            await sleep_ms(t)


async def retry(n, t, *a, **k):
    "call a function every @t seconds until it succeeds"
    return retry_ms(n, int(t * 1000), *a, **k)


if DEBUG:

    async def _catch(n, p, *a, **k):
        try:
            print("RRR", n, file=sys.stderr)
            return await p(*a, **k)
        except Exception as exc:
            log("Error: %s %r", n, exc, err=exc)
            raise
        else:
            print("Done:", n, file=sys.stderr)


class TaskGroup(_tg):
    "anyio.TaskGroup, lightly enhanced"

    async def spawn(self, p, *a, _name=None, **k):
        "Starts a task now. Returns something you can cancel."
        if DEBUG:
            print("RUN", _name, p, a, k, file=sys.stderr)
            return self.create_task(_catch(_name, p, *a, **k))  # , name=_name)
        else:
            return self.create_task(p(*a, **k))  # , name=_name)

    def start_soon(self, p, *a, _name=None, **k):
        "Starts a task soon."
        if DEBUG:
            print("RUN", _name, p, a, k, file=sys.stderr)
            self.create_task(_catch(_name, p, *a, **k))
        else:
            self.create_task(p(*a, **k))


def run(p, *a, **k):
    "like anyio.run"
    return _run(p(*a, **k))


# Helper task to run a TCP stream server.
# Callbacks (i.e. connection handlers) may run in a different taskgroup.
async def run_server(cb, host, port, backlog=5, taskgroup=None, evt=None) -> Never:
    """
    Task that runs a TCP stream server.
    Callbacks (i.e. connection handlers) may run in a different taskgroup.

    The optional event is set when the socket is listening.
    """
    import socket  # noqa: PLC0415

    # Create and bind server socket.
    host = socket.getaddrinfo(host, port)[0]  # TODO this is blocking!
    s = socket.socket()
    s.setblocking(False)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(host[-1])
    s.listen(backlog)
    if evt is not None:
        evt.set()

    try:
        if taskgroup is None:
            async with TaskGroup() as tg:
                await _run_server(tg, s, cb)
        else:
            await _run_server(taskgroup, s, cb)
    finally:
        s.close()


async def _run_server(tg, s, cb):
    from asyncio import core as _core  # noqa: PLC0415

    while True:
        if DEBUG:
            print("WaitServer", file=sys.stderr)
        yield _core._io_queue.queue_read(s)  # noqa:SLF001
        if DEBUG:
            print("WaitedServer", file=sys.stderr)
        try:
            s2, addr = s.accept()
        except Exception as err:
            # Ignore a failed accept
            print("ErrServer", repr(err), file=sys.stderr)
            continue

        if DEBUG:
            print("GotServer", file=sys.stderr)
        s2.setblocking(False)
        # XXX uasyncio implementation detail
        s2s = asyncio.StreamReader(s2, {"peername": addr})
        await tg.spawn(cb, s2s)


# minimal Outcome clone


class Outcome:
    "a minimal stub for `outcome.Outcome`"

    def __init__(self, val):
        self.val = val


class Value(Outcome):
    "a wrapped non-error value"

    def unwrap(self):
        "unwrap and return the value"
        try:
            return self.val
        finally:
            del self.val


class Error(Outcome):
    "a wrapped exception"

    def unwrap(self):
        "unwrap and raise the exception"
        try:
            raise self.val
        finally:
            del self.val


class ValueEvent:
    """
    A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and a task can wait for it to become True.

    Note that the value can only be read once.
    """

    def __init__(self, scope=None):
        self.event = Event()
        self.value = None
        self.scope = scope

    def set(self, value):
        """
        Set the result to return this value, and wake any waiting task.
        """
        self.value = Value(value)
        self.event.set()

    def set_error(self, exc):
        """
        Set the result to raise this exception, and wake any waiting task.
        """
        self.value = Error(exc)
        self.event.set()

    def is_set(self):
        """
        Check whether the event has occurred.
        """
        return self.value is not None

    def cancel(self):
        """
        Send a cancelation to the recipient.
        """
        if self.scope is not None:
            self.scope.cancel()
        self.set_error(CancelledError())

    async def get(self):
        """
        Block until the value is set.

        If it's already set, then this method returns immediately.

        The value can only be read once.
        """
        await self.event.wait()
        return self.value.unwrap()


class aclosing:
    """contextlib.aclosing"""

    def __init__(self, thing):
        self.thing = thing

    async def __aenter__(self):
        return self.thing

    async def __aexit__(self, *exc_info):
        await self.thing.aclose()


def ACM(obj):
    """A bare-bones async context manager.

    Usage::

        class Foo():
            async def __aenter__(self):
                AC = ACM(obj)
                ctx1 = await AC(obj1)
                ...
            async def __aexit__(self, *exc):
                return await AC_exit(self, *exc)

    Calls to `ACM` and `AC_exit` can be nested. They **must** balance.
    """

    if hasattr(obj, "_AC_"):
        obj._AC_.append(None)
    else:
        obj._AC_ = []

    def _ACc(ctx):
        return AC_use(obj, ctx)

    return _ACc


async def AC_use(obj, ctx):
    """
    Attach a callback / (async) context manager to this object's AC.
    """
    if hasattr(ctx, "__aenter__"):
        cm = await ctx.__aenter__()
    elif hasattr(ctx, "__enter__"):
        cm = ctx.__enter__()
    else:
        cm = None
    obj._AC_.append(ctx)
    return cm


async def AC_exit(obj, *exc):
    """End the latest async context manager opened by `ACM`."""
    received_exc = exc[0] is not None

    # Callbacks are invoked in LIFO order to match the behaviour of
    # nested context managers.
    if not exc:
        exc = (None, None, None)

    suppressed_exc = False
    pending_raise = False
    while obj._AC_:
        cb = obj._AC_.pop()
        if cb is None:
            break
        try:
            if hasattr(cb, "__aexit__"):
                res = await cb.__aexit__(*exc)
            elif hasattr(cb, "__exit__"):
                res = cb.__exit__(*exc)
            else:
                res = cb()
                if hasattr(res, "throw"):  # async cb
                    res = await res

            if res:  # suppress error
                suppressed_exc = True
                pending_raise = False
                exc = (None, None, None)
        except BaseException as ex:
            exc = (type(ex), ex, getattr(ex, "__traceback__", None))
            pending_raise = True
    if pending_raise:
        raise exc[1]
    return received_exc and suppressed_exc


class _Shield:
    def __enter__(self):
        pass

    def __exit__(self, *tb):
        pass


_shield = _Shield()


def shield():
    "no-op context manager, supposed to shield a (sub)task from cancels"
    return _shield


# "await" for stream read/write-ability

from asyncio import core  # noqa:E402


def _rdq(s):  # async
    yield core._io_queue.queue_read(s)  # noqa:SLF001


def _wrq(s):  # async
    yield core._io_queue.queue_write(s)  # noqa:SLF001


def is_async(obj):  # noqa: D103
    if iscoroutine(obj) or hasattr(obj, "__iter__"):
        return True
    return False


class _Thr:
    def __init__(self, p, a, k):
        self.p = p
        self.a = a
        self.k = k
        self.evt = asyncio.ThreadSafeFlag()
        self.res = None
        self.err = None

    def run(self):
        try:
            self.res = self.p(*self.a, **self.k)
        except BaseException as exc:
            self.err = exc
        self.evt.set()


if sys.platform == "rp2":
    # RP2 only can do a single thread at a time
    _thr_lock = Lock()
else:

    class _DummyLock:
        async def __aenter__(self):
            return

        async def __aexit__(self, *tb):
            return

    _thr_lock = _DummyLock()


async def to_thread(p, *a, **k):
    """Run this function in a thread.

    Some platforms only have one thread (e.g. RP2). Thus don't do
    long-running things in there.

    Also startup cost is not trivial due to result and exception propagation.
    """
    thr = _Thr(p, a, k)
    import _thread  # noqa:PLC0415

    async with _thr_lock:
        _thread.start_new_thread(thr.run, ())
        await thr.evt.wait()
    if thr.err is not None:
        raise thr.err from None
    return thr.res



"""
Apps used for testing.
"""

from __future__ import annotations

import sys

from moat.lib.micro import Event, L, Queue, log, sleep_ms, wait_for_ms
from moat.lib.proxy import as_proxy
from moat.lib.rpc import BaseCmd, NoStream

# Typing
from typing import TYPE_CHECKING  # isort:skip


@as_proxy("_T_UCr")
class UserCrash(RuntimeError):
    pass


if TYPE_CHECKING:
    from moat.lib.rpc import Msg

    from collections.abc import Awaitable
    from typing import Any


class Cmd(BaseCmd):
    """
    A rather basic test command.
    """

    doc = dict(_c=dict(_d="Basic test cmd"))

    n = 0
    err: Exception
    err_evt: Event
    store: list[Any]

    doc_echo = dict(_d="Echo. Returns 'm'", m="any", _r=dict(r="any:m"))

    async def cmd_echo(self, m: Any):
        "Basic echo method, returns @m as ``result['r']``"
        return {"r": m}

    doc_it = dict(_d="Iterator. Sends 0…lim-1.", lim="int:limit", _o="int", delay="float:timer")

    async def stream_it(self, msg: Msg):
        "Streams numbers."
        lim = msg.get("lim", -1)
        i = 0
        d = int(msg.get("delay", 0.1) * 1000)
        async with msg.stream_out() as s:
            while i != lim:
                await sleep_ms(d)
                try:
                    await s.send(i)
                except NoStream:
                    break
                i += 1

    doc_nit = dict(_d="Call counter.", lim="int:limit", delay="float:timer")

    async def cmd_nit(self, delay: float = 0):
        "A non-iterator counter; simply counts calls to it."
        self.n += 1
        d = int(delay * 1000)
        await sleep_ms(d)
        return self.n

    doc_store = dict(_d="Save data.", _0="any:Data", _r="list[any]:collected data if no param")

    async def stream_store(self, msg):
        "Collect data (or return them)"
        if msg.args:
            self.store.extend(msg.args)
            st = ()
        else:
            st, self.store = self.store, []
        await msg.result(*st)

    doc_clr = dict(_d="Clear the counter.", n="int:new value, default zero")

    async def cmd_clr(self, n: int = 0):
        self.n = n

    doc_print = dict(_d="print string", _0="str:output", e="bool:use stderr")

    async def cmd_print(self, d: str, e: bool = False):
        "print to stdout/stderr"
        print(d, file=sys.stderr if e else sys.stdout)

    doc_error = dict(_d="raise exc", e="exc:raised")

    async def cmd_error(self, e: Exception | type[Exception] = RuntimeError):
        "raise an exception"
        if isinstance(e, Exception):
            raise e
        else:
            raise UserCrash(e)

    doc_crash = dict(_d="cause a crash", e="exc:raised")

    async def cmd_crash(self, e: Exception | type[Exception] = UserCrash, *a):
        "crash this app"
        if isinstance(e, Exception):
            assert not a
            self.err = e
        elif isinstance(e, type):
            assert issubclass(e, Exception)
            self.err = e(*a)  # the remote might send a
        else:
            self.err = UserCrash(e, *a)  # the remote might send a text instead
        self.err_evt.set()

    async def stream_run(self, msg: Msg):
        """
        Originate a command.
        """
        if msg.can_stream:
            raise NotImplementedError
        res = await self.root.sender.cmd(msg.args[0], *msg.args[1:], **msg.kw)
        await msg.result(*res.args, **res.kw)

    async def setup(self):
        self.err_evt = Event()
        self.store = []
        await super().setup()

    async def task(self):
        """
        Idle task, except that it crashes the app 100 ms after *crash* is called.
        """
        if L:
            self.set_ready()
        await self.err_evt.wait()
        await sleep_ms(100)
        raise self.err


class Cons(BaseCmd):
    """
    A console reader.

    Config:

        cons: path to the *Blk / *Msg reading the data
        lines: max lines to queue up
        prefix: log prefix.

    If the prefix is empty, data is queued up for reading via :meth:`cmd_rd`.
    """

    doc = dict(
        _c=dict(
            _d="Basic console reader",
            prefix="str:log prefix, empty=queued for 'rd'",
            cons="path:obj with crd",
            lines="int:queue len",
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)

    async def setup(self):
        await super().setup()
        self.con = self.root.sub_at(self.cfg["cons"])
        if self.cfg.get("prefix", None) is None:
            self.q = Queue(self.cfg.get("lines", 10))
        self.timeout = self.cfg.get("timeout", 200)

    doc_rd = dict(_d="read console data", _r="str:one line")

    def cmd_rd(self) -> Awaitable:
        return self.q.get()

    async def task(self):
        if L:
            self.set_ready()
        await self.con.cmd("rdy_")

        buf = bytearray(200)
        d = 0
        while True:
            timed = False
            try:
                if d:
                    b = await wait_for_ms(self.timeout, self.con.crd, n=len(buf) - d)
                else:
                    b = await self.con.crd(n=len(buf))
            except TimeoutError:
                timed = True
            else:
                buf[d : d + len(b)] = b
                d += len(b)
            if d == len(buf) or (d > 0 and (timed or buf[d - 1] == 10)):  # lf
                p = self.cfg.get("prefix", None)
                if p is None:
                    self.q.put(buf[:d])
                else:
                    log(
                        "%s: %s",
                        p,
                        bytes(memoryview(buf)[: d - (buf[d - 1] == 10)]).decode("utf-8"),
                    )
                d = 0



"""
Communicate with diyBMS (MoaT firmware)
"""

from __future__ import annotations


def Comm(cfg):
    """
    Communicator for the serially-connected cell controllers.

    This app accepts control packets, encodes and forwards them
    to the link, and returns the reply packets.
    """
    from moat.ems.battery.diy_serial.comm import BattComm  # noqa: PLC0415

    return BattComm(cfg)


def Cell(cfg):
    """
    Direct interface to a single cell.

    @comm: BattComm instance
    @i: cell number there

    This BaseCell translates commands to Comm requests.
    """
    from moat.ems.battery.diy_serial.cell import Cell  # noqa: PLC0415

    return Cell(cfg)



"""
Adaptor for MicroPython streams.
"""

from __future__ import annotations

from moat.lib.micro import TimeoutError, _rdq, _wrq, wait_for_ms  # noqa:A004
from moat.lib.stream import BaseBuf


class FileBuf(BaseBuf):
    """
    Bytestream > sync MicroPython stream

    Reads a byte at a time if the stream doesn't have an "any()" method.

    @timeout times out short reads if no more data arrives, if >0.

    @force_write must be set if the write side doesn't support polling.

    Override the `setup` async context manager to set up and tear down the
    stream. It must yield either a single file or a stdin/stdout tuple.
    """

    _buf = None
    _any = lambda: 1  # noqa:E731

    def __init__(self, cfg: dict | None = None, force_write=False, timeout=2):
        super().__init__(cfg or {})
        self.force_write = force_write
        self.timeout = timeout

    async def setup(self):  # noqa:D102
        s = await self.stream()
        if isinstance(s, tuple):
            self.rs, self.ws = s
        else:
            self.rs = self.ws = s
        self._any = getattr(self.rs, "any", lambda: 1)

    async def stream(self):  # noqa:D102
        raise NotImplementedError

    async def rd(self, buf):
        "forwards to ``.read(into)``"
        n = 0
        m = memoryview(buf)
        while m:
            if n == 0 or self.timeout is None:
                await _rdq(self.rs)
            else:
                try:
                    await wait_for_ms(self.timeout, _rdq, self.rs)
                except TimeoutError:
                    break
            d = self.rs.readinto(m[: min(self._any(), len(m))])
            if not d:
                break
            m = m[d:]
            n += d
        return n

    async def wr(self, buf):
        "forwards to ``.write``, handles short writes"
        buf = memoryview(buf)
        t = len(buf)
        while buf:
            if not self.force_write:  # XXX *sigh*
                await _wrq(self.ws)
            n = self.ws.write(buf)
            if n is None:
                n = len(buf)
            buf = buf[n:]
        return t



"""
Apps used for structure.
"""

from __future__ import annotations

from typing import TYPE_CHECKING as _TC

if _TC:
    from .array import Array as Array
    from .dir import Dir as Dir
    from .err import Err as Err

_imports = {
    "Array": "array",
    "Dir": "dir",
    "Err": "err",
}


def __getattr__(attr: str):
    try:
        mod = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, (attr,), 1), attr)
    globals()[attr] = value
    return value


def __dir__():
    """Expose all lazy-loaded attributes for introspection."""
    return sorted(set(list(__all__) + [k for k in globals().keys() if not k.startswith("_")]))


__all__ = ["Array", "Dir", "Err"]



"""
Stream link-up support for MoaT commands
"""

from __future__ import annotations

from moat.lib.stream import Base, BaseBlk, BaseBuf, BaseMsg

__all__ = ["BlkCmd", "BufCmd", "MsgCmd"]

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.stream.base import Buffer, MutBuffer

    from typing import Any


class BBMCmd(Base):
    """
    Generic base for *Cmd.
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        self.cmd = cfg._moat_cmd  # noqa:SLF001

    async def setup(self):
        await Base.setup(self)
        # not using super() because {Msg,Buf,Base}Cmd pull in inheritance
        # from BaseConn which calls ``.stream`` which we don't have, or want
        self.s = self.cmd.root.sub_at(self.cfg["path"])


class MsgCmd(BBMCmd, BaseMsg):
    """
    A stream handler that forwards send/recv (and console) requests via MoaT.

    This is the reverse of a CmdBBM for messages, i.e. a stream handler that forwards
    send/recv (and console) requests via MoaT.

    The remote link is addressed by the config item "path".
    """

    # pylint:disable=abstract-method

    async def send(self, m: Any) -> Any:  # pylint:disable=invalid-overridden-method
        "send a message"
        return await self.s.s(m=m)

    async def recv(self) -> Any:  # pylint:disable=invalid-overridden-method
        "receive a message"
        return await self.s.r()

    async def cwr(self, buf: Buffer) -> None:  # pylint:disable=invalid-overridden-method
        "write console data"
        await self.s.cwr(b=buf)

    async def crd(self, buf: MutBuffer) -> int:
        "read console data"
        msg = await self.s.crd(n=len(buf))
        buf[: len(msg)] = msg
        return len(msg)


class BufCmd(BBMCmd, BaseBuf):
    """
    A stream handler that forwards snd/rcv (and console) requests via MoaT.

    This is the reverse of a CmdBBM for buffers, i.e. a stream handler that forwards
    snd/rcv (and console) requests via MoaT.

    The remote link is addressed by the config item "path".
    """

    # pylint:disable=abstract-method
    # `stream` needs to be implemented by a subclass

    async def wr(self, data: Buffer) -> int:  # noqa:D102
        # pylint: disable=invalid-overridden-method
        return await self.s.wr(data)

    async def rd(self, buf: MutBuffer) -> int:  # noqa:D102
        msg = await self.s.rd(n=len(buf))
        buf[: len(msg)] = msg
        return len(msg)


class BlkCmd(BBMCmd, BaseBlk):
    """
    A stream handler that forwards snd/rcv (and console) requests via MoaT.

    This is the reverse of a CmdBBM for blocks, i.e. a stream handler that forwards
    snd/rcv (and console) requests via MoaT.

    The remote link is addressed by the config item "path".
    """

    # pylint:disable=abstract-method

    crd = MsgCmd.crd
    cwr = MsgCmd.cwr

    async def snd(self, m: Buffer | bytes) -> None:  # noqa:D102
        # pylint: disable=invalid-overridden-method
        await self.s.sb(m=m)

    async def rcv(self) -> Buffer | bytes:  # noqa:D102
        # pylint: disable=invalid-overridden-method
        return await self.s.rb()



"""
the WDT doesn't need a server-side command handler
"""

from __future__ import annotations

from moat.lib.rpc import BaseCmd


class Cmd(BaseCmd):
    "empty"

    doc = dict(_c=dict(_d="dummy watchdog"))
    # pylint:disable=unnecessary-pass



"""
MoaT satellite boot script.

This file adds some essential setup.

Local code should go in "boot_local.py".
This file may be overwritten by a MoaT update.
"""

# from moat import setup
# setup.run()
from __future__ import annotations

import os
import sys

if sys.platform == "linux":
    if (_t := int(os.getenv("MOAT_FORK_DEBUG", "0"))) > 0:
        from time import sleep as _sleep

        _sleep(_t)
        del _sleep
    del _t
else:
    try:
        sys.path.remove("/lib")
    except ValueError:
        pass
    sys.path.insert(0, "/lib")

import moat  # just for the namespace

if not hasattr(moat, "SERIAL"):
    try:
        import usb.device
        from usb.device.cdc import CDCInterface

        moat.SERIAL = CDCInterface()
        try:
            moat.SERIAL.init(timeout=0)
            usb.device.get().init(moat.SERIAL, builtin_driver=True)
        except Exception:
            del moat.SERIAL
            raise
        finally:
            del CDCInterface
            del usb

    except Exception as exc:
        moat.SERIAL_EXC = exc

print("\n*** MoaT ***\n", file=sys.stderr)

# One line for handling boot_local, ten for not doing it. Fits.
try:
    from boot_local import *  # noqa:F403
except Exception as exc:  # noqa:F841  # SIGH
    for ext in ("py", "mpy"):
        try:
            os.stat(f"boot_local.{ext}")
        except OSError:
            pass
        else:
            exc = None
            break
    if exc is not None and (not isinstance(exc, ImportError) or "boot_local" not in repr(exc)):
        sys.print_exception(exc)



# noqa:D104 pylint:disable=missing-module-docstring
from __future__ import annotations

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from moat.lib.config import register as _register
from moat.lib.rpc import add_app_prefix

_register(__name__)
add_app_prefix("moat.micro.app")



"""
Console class.

Set up from the root "main.py".
"""

from __future__ import annotations

import io
import os

from asyncio import create_task, run_until_complete

from moat.lib.micro import (
    ACM,
    AC_exit,
    AC_use,
    CancelledError,
    Event,
    TaskGroup,
    TimeoutError,  # noqa:A004
    ValueEvent,
    wait_for_ms,
)
from moat.lib.ring import RingBuffer
from moat.lib.ring.aio import RingBuffer as AioRingBuffer
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from asyncio import Task

    from moat.lib.rpc import Msg

    from collections.abc import Awaitable, Callable

#
# The purpose of this code is to run asyncio from the console read loop.
#
# The main task (moat.micro.main) creates the Main class, starts the
# dispatch task, and saves it in `main.main_task`. The context frames are
# also attached to the `main` object.
#
# Closing the console cancels the main task and unwinds the contexts.
# After that the system should be in the same state it would be in if MoaT
# had ended without running the REPL behind the scenes (or rather, running
# the REPL while processing MoaT behind the scenes).


class Main:
    """
    This is a singleton-ish class that holds the main execution context
    and (possibly) the console reference.

    This class is not standalone. It is created in `./main.py`.
    """

    console: Console | None = None
    main_task: Task = None
    user_task: Task = None
    tg: TaskGroup = None  # created in ./main.py

    def __init__(self, wr_exc: Callable):
        ACM(self)

        self.evt = evt = Event()
        self.wait = evt.wait
        self.wr_exc = wr_exc

    def start(self, cons: Console):
        """
        Start using this console.

        Called from `Console.start`.
        """
        import moat  # noqa:PLC0415

        moat.call = call
        moat.bg = bg

        self.console = cons
        self.evt.set()

    def stop(self):
        """
        Stop using the console, if any.

        Called from the MoaT main code.
        """
        if self.console is not None:
            import moat  # noqa:PLC0415

            moat.call = None
            if self.main_task is not None:
                self.main_task.cancel()
                self.main_task = None
            try:
                self.console.stop()
            finally:
                self.console = None

    def die(self, exc: Exception):
        """
        Stop using the console and unwind.

        Called from the MoaT main code.
        """
        run_until_complete(create_task(self.die_(exc)))

    async def die_(self, exc: Exception):
        "async version of `die`."
        self.stop()
        if not isinstance(exc, CancelledError):
            self.wr_exc(exc)
        while self._AC_:
            await AC_exit(self, type(exc), exc, None)


main: Main = None  # set by moat.micro.main


def bg(p, *a, **k) -> ValueEvent:
    """
    Returns a ValueEvent that encapsulates the result of running @p,
    and which allows it to be cancelled.

    Installed as 'moat.bg'.
    """
    evt = ValueEvent()

    async def run_(evt, p, a, k):
        try:
            evt.set(await p(*a, **k))
        except BaseException as exc:
            evt.set(exc)

    async def run(evt, p, a, k):
        evt.scope = await main.tg.spawn(run_, evt, p, a, k)
        return evt

    run_until_complete(create_task(run(evt, p, a, k)))
    return evt


def call(p, *a, **k):
    """
    Returns the result of running ``p`` asynchronously.

    A REPL input that consists of a single Ctrl-C will terminate the call.

    Installed as 'moat.call'.
    """
    if main.user_task is not None:
        raise RuntimeError("A user task is already running!")
    main.user_task = evt = bg(p, *a, **k)
    try:
        return run_until_complete(evt.get())
    finally:
        main.user_task = None


class Console(io.IOBase):
    """
    The console driver that allows MoaT to run in the background.

    Parameters:
        keep (bool): flag whether to keep the original console driver, required
        sleep (int): timeout for asyncio; milliseconds. Default: 20
        repl (bool): flag whether to 'return' to the REPL. Default: `False`.
    """

    _term = None
    _wr_evt: Event | None = None

    def __init__(self, cfg):
        self.cfg = cfg
        self._AC_ = main._AC_

        self._rb = AioRingBuffer(128)
        self._wb = RingBuffer(512)

    def write(self, buf: bytes) -> int:
        "Console write"
        if not buf:
            return 0
        if self._wr_evt is not None:
            self._wr_evt.set()
        return self._wb.write(buf)

    async def get_out(self, buf: bytearray) -> int:
        "extract from write buffer"
        while not len(self._wb):
            self._wr_evt = Event()
            try:
                await self._wr_evt.wait()
            finally:
                self._wr_evt = None
        return self._wb.readinto(buf)

    def readinto(self, buf: bytearray):
        "Console readinto, runs the asyncio loop"
        try:
            res = run_until_complete(
                create_task(wait_for_ms(self.cfg.get("sleep", 20), self._rb.readinto, buf))
            )
        except TimeoutError:
            # if self._wb.n_avail < 100: self.write(b"_")
            return None
        except BaseException as exc:
            self.__exit__()
            main.die(exc)
            run_until_complete(AC_exit(self, type(exc), exc, None))
            raise
        else:
            return res

    def read(self, n):
        "Console read, runs the asyncio loop"
        buf = bytearray(n)
        n = self.readinto(buf)
        if not n:
            return None
        if n < len(buf):
            buf[n:] = b""
        return buf

    def ioctl(self, req, flags):
        """
        ioctl to test for readiness.
        """
        # if self._wb.n_avail < 100: self.write(f"IOC {req} {flags}\n".encode("utf-8"))
        if req == 3:  # MP_STREAM_POLL
            if flags & 1:
                try:
                    if not len(self._rb):
                        run_until_complete(
                            create_task(
                                wait_for_ms(self.cfg.get("sleep", 20), self._rb.wait_avail)
                            )
                        )
                except TimeoutError:
                    # if self._wb.n_avail < 100: self.write(b"-")
                    flags &= ~1

            return flags & 5  # read or write

        # TODO 4 is MP_STREAM_CLOSE

        return -1  # Other requests are unsupported

    async def put_in(self, buf: bytes | memoryview) -> Awaitable:
        "feed to read buffer"
        if main.user_task is not None and len(buf) == 1 and buf[0] == 3:
            main.user_task.cancel()
        await self._rb.write(buf)

    def start(self):
        "Start using this."
        if main.console is not None:
            raise RuntimeError("Already up")

        # if self._wb.n_avail < 100: self.write(b"Dup ")
        if self.cfg["keep"]:
            os.dupterm(self, 1)
        else:
            self._term = os.dupterm(self, 0)
        # if self._wb.n_avail < 100: self.write(b"Duped.\n")
        if self.cfg.get("repl", False):
            main.start(self)
        return self

    __enter__ = start

    def stop(self):
        "Stop using this."
        if main.console is self:
            main.console = None
        if self.cfg["keep"]:
            os.dupterm(None, 1)
        else:
            os.dupterm(self._term, 0)
            self._term = None

    def __exit__(self, *exc):
        self.stop()


class Cmd(BaseCmd):
    """
    The command that attaches and runs a console.

    See :cls:`Console` for configuration.
    """

    async def setup(self):
        "Create a console"
        await super().setup()
        self.cons = await AC_use(self, Console(self.cfg))

    doc_r = dict(_d="Read console data", _0="int:maxlength (default 32)")

    async def cmd_r(self, n: int = 32) -> bytes:
        "read from stdout/stderr"
        buf = bytearray(n)
        res = await self.cons.get_out(buf)
        if not res:
            buf = b""
        elif res < n:
            buf = memoryview(buf)[:res]
        return buf

    doc_w = dict(_d="Write console data", _0="bytes: data")

    async def cmd_w(self, data: bytes):
        "write stdin"
        await self.cons.put_in(data)

    doc_rw = dict(
        _d="r/w console byte stream",
        _0="int:rdbuflen (64)",
        _i="bytes:for stdin",
        _o="bytes:from stdout+stderr",
    )

    async def stream_rw(self, msg):
        "read/write the console stream"
        n = msg.get(0, 32)
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def rd():
                while True:
                    await st.send(await self.cmd_r(n))

            async for m in st:
                await self.cmd_w(m[0])
            tg.cancel()

    doc_rb = dict(_d="Read input buffer", _0="int:maxlength (default 32)")

    async def stream_rb(self, msg: Msg):
        "read from stdin"
        buf = bytearray(msg.get(0, 32))
        if msg.can_stream:
            async with msg.stream_w() as ms:
                while True:
                    res = await self.cons.readinto(buf)
                    await ms.send(buf[:res])
        else:
            res = await self.cons.readinto(buf)
            if not res:
                buf = b""
            else:
                buf = memoryview(buf)[:res]
            await msg.result(buf)



"""
This is a button in/out multiplexer.
"""

from __future__ import annotations

import asyncio
from machine import Pin, Timer, disable_irq, enable_irq
from micropython import const

from moat.util import merge

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

DEFAULT = dict(
    input=(),
    output=(),
    inv=dict(scan=False, input=False, output=False),
)

SCAN = const(0)
INPUT = const(1)
OUTPUT = const(2)


class Multiplex:
    """
    Run a pin multiplexer.

    The driver pins are a list of pin numbers at cfg.scan.
    The inputs are connected to cfg.input, while the outputs are on cfg.output.
    Inputs or outputs may be empty.
    """

    def __init__(self, cfg, timer=None):
        merge(cfg, DEFAULT, replace=False)
        self.cfg = cfg
        inv = cfg["inv"]
        self.inv = inv = [inv["scan"], inv["input"], inv["output"]]  # see constants!

        self.scan = [Pin(x, mode=Pin.OUT, value=inv[SCAN]) for x in cfg["scan"]]
        self.input = [Pin(x, mode=Pin.IN, value=inv[INPUT]) for x in cfg["input"]]
        self.output = [Pin(x, mode=Pin.OUT, value=inv[OUTPUT]) for x in cfg["output"]]

        self.this = None

        self.state = 0
        self.changed = 0
        self.work = asyncio.ThreadSafeFlag()
        self.out = 0  # output state

        if timer is None:
            timer = cfg["timer"]
            if isinstance(timer, int):
                timer = Timer(timer)
        self.timer = timer

    def __getitem__(self, btn: int) -> bool:
        "Read input pin state"
        return bool(self.state & (1 << btn))

    get_in = __getitem__

    def __setitem__(self, btn: int, value: bool) -> None:
        "Write output pin state"
        mask = 1 << btn
        self.out = (self.out & ~mask) | (mask if value else 0)

    set_out = __setitem__

    def get_out(self, btn: int) -> bool:
        "Read output pin state"
        return bool(self.out & (1 << btn))

    def changes(self) -> Iterator[tuple[int, int]]:
        """
        Iterator that yields (button, true-if-pushed) tuples.
        """
        btn = len(self.scan) * len(self.input)
        if btn == 0:
            return

        self.work.clear()
        btn -= 1
        mask = 1 << btn
        while btn >= 0:
            if self.changed & mask:
                i_ = disable_irq()
                self.changed &= ~mask
                enable_irq(i_)
                yield btn, bool(self.state & mask)
            mask >>= 1
            btn -= 1

    def start(self):
        "Start the multiplexer"
        if self.this is not None:
            raise RuntimeError("already running")
        self.this = 0
        self.imask = self.omask = 1
        self.step()
        self.timer.init(
            mode=Timer.PERIODIC, period=self.cfg["msec"], callback=self.step, hard=False
        )

    def stop(self):
        "Stop the multiplexer, de-assert all pins"
        self.timer.deinit()
        self.this = None

        v = self.inv[OUTPUT]
        for pin in self.output:
            pin(v)

        v = self.inv[SCAN]
        for pin in self.scan:
            pin(v)

    def step(self, _timer=None):
        """Run one pass of the multiplexer"""
        vs = self.inv[SCAN]
        vi = self.inv[INPUT]
        vo = self.inv[OUTPUT]

        this = self.this
        imask = self.imask
        omask = self.omask

        # Read inputs.
        for pin in self.input:
            val = 0 if pin() == vi else imask
            if (self.state & imask) != val:
                self.state = (self.state & ~imask) | val
                self.changed |= imask
                self.work.set()
            imask <<= 1

        # Turn off old outputs.
        # Could only turn off those that won't stay on
        # but that eats time.
        for pin in self.output:
            pin(vo)

        self.scan[this](vs)

        this += 1
        if this == len(self.scan):
            this = 0
            imask = omask = 1

        self.scan[this](not vs)
        # turn on new outputs
        for pin in self.output:
            if self.out & omask:
                pin(not vo)
            omask <<= 1
        self.this = this
        self.imask = imask
        self.omask = omask



"""
This module contains an async version of mpy_repl.repl_connection.MpyPath,
and an equivalent that uses the MoaT file access protocol

"""

# CPython only

from __future__ import annotations

import anyio
import binascii
import hashlib
import io
import logging
import os
import stat
import sys
from contextlib import suppress
from os import PathLike
from pathlib import Path

from moat.util import attrdict
from moat.lib.codec.errors import RemoteError
from moat.util.exec import CalledProcessError, run

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from typing import Self

logger = logging.getLogger(__name__)


class ABytes(io.BytesIO):
    """
    An async-IO-mimicing version of `io.BytesIO`.
    """

    def __init__(self, name, data):
        super().__init__()
        self.name = name
        self.write(data)
        self.suffix = Path(name).suffix

    def __str__(self):
        return str(self.name)

    async def open(self, mode):  # noqa:ARG002 pylint: disable=unused-argument
        "reset the buffer pointer"
        self.seek(0, 0)
        return self

    async def read_bytes(self):
        "return the current buffer"
        return self.getbuffer()

    async def sha256(self, l: int | None = None):  # noqa: E741
        "hash the current buffer"
        _h = hashlib.sha256()
        _h.update(self.getbuffer())
        res = _h.digest()
        if l is not None:
            res = res[:l]
        return res

    async def gitsha1(self):
        "hash the current buffer"
        _h = hashlib.sha1()  # noqa:S324  # sha1
        _h.update(f"blob {len(self.getbuffer())}\0".encode())
        _h.update(self.getbuffer())
        return _h.digest()

    def close(self):
        "does nothing"
        # pylint:disable=unnecessary-pass

    async def is_dir(self):
        "returns False"
        return False

    async def is_file(self):
        "returns True"
        return True

    async def stat(self, follow_symlinks=False):  # noqa:ARG002 pylint: disable=unused-argument
        "returns the size. All other stat fields are empty"
        res = attrdict()
        res.st_size = len(self.getbuffer())
        return res


class APath(anyio.Path):
    "anyio.Path, enhanced with a couple of our methods"

    def with_suffix(self, suffix: str) -> APath:
        """Return a copy with a changed suffix."""
        return APath(super().with_suffix(suffix))

    def joinpath(self, *args: str | PathLike[str]) -> APath:
        """Join path components and keep :class:`APath`."""
        return APath(super().joinpath(*args))

    @property
    def parent(self) -> APath:
        """Parent directory as :class:`APath`."""
        return APath(super().parent)

    async def sha256(self, l=None) -> bytes:  # noqa: E741
        """
        :returns: hash over file contents

        Calculate a SHA256 over the file contents and return the digest.
        """
        _h = hashlib.sha256()
        async for block in self.read_as_stream():
            _h.update(block)
        res = _h.digest()
        if l is not None:
            res = res[:l]
        return res

    async def gitsha1(self):
        """
        :returns: hash over file contents

        Calculate a SHA1 over the file contents, the way "git" does it.
        """
        _h = hashlib.sha1()  # noqa:S324  # sha1
        sz = (await self.stat()).st_size
        _h.update(f"blob {sz}\0".encode())
        async for block in self.read_as_stream():
            _h.update(block)
        return _h.digest()

    async def read_as_stream(self, chunk=4096):  # noqa:ARG002
        """
        :returns: async Iterator
        :rtype: Iterator of bytes

        Iterate over blocks (`bytes`) of a remote file.
        """
        # pylint:disable=unused-argument

        try:
            d = await self.read_bytes()
            yield d
        except EOFError:
            pass

    async def iterdir(self) -> AsyncIterator[APath]:
        """Yield child paths as :class:`APath` objects."""
        async for p in super().iterdir():
            yield APath(p)


class MoatPath(anyio.Path):  # pathlib.PosixPath
    "abstract superclass for MoaT-connected async file access"

    _stat_cache: os.stat_result | None = None
    _repl: object | None = None

    def connect_repl(self, repl: object) -> Self:
        """Connect object to remote connection."""
        self._repl = repl
        return self  # allow method joining

    def _with_stat(self, st: os.stat_result) -> Self:
        self._stat_cache = os.stat_result(st)
        return self

    # methods to override to connect to repl

    def with_name(self, name: str) -> Self:
        "set name"
        # pylint: disable=no-member
        return type(self)(super().with_name(name)).connect_repl(self._repl)

    def with_suffix(self, suffix: str) -> Self:
        "set suffix"
        # pylint: disable=no-member
        return type(self)(super().with_suffix(suffix)).connect_repl(self._repl)

    def relative_to(self, *other: str | PathLike[str]) -> Self:
        "return relative path"
        # pylint: disable=no-member
        return type(self)(super().relative_to(*other)).connect_repl(self._repl)

    def joinpath(self, *args: str | PathLike[str]) -> Self:
        "join paths"
        # pylint: disable=no-member
        return type(self)(super().joinpath(*args)).connect_repl(self._repl)

    @property
    def parent(self):
        "parent directory"
        # pylint: disable=no-member
        return type(self)(super().parent).connect_repl(self._repl)

    async def glob(
        self,
        pattern: str | PathLike[str],
        *,
        case_sensitive: bool | None = None,  # noqa:ARG002
        recurse_symlinks: bool = False,  # noqa:ARG002
    ):
        """
        :param str pattern: string with optional wildcards.
        :return: generator over matches (MpyPath objects)

        Pattern match files on remote.
        """
        pattern = str(pattern).removeprefix("/")  # XXX
        parts = pattern.split("/")
        # print('glob', self, pattern, parts)
        if not parts:
            return
        elif len(parts) == 1:
            async for r in self.iterdir():
                if r.match(pattern):
                    yield r
        else:
            remaining_parts = "/".join(parts[1:])
            # pylint:disable=no-else-raise
            if parts[0] == "**":
                raise NotImplementedError
                # for dirpath, dirnames, filenames in walk(self):
                # for path in filenames:
                # if path.match(remaining_parts):
                # yield path
            else:
                async for path in self.iterdir():
                    if (await path.is_dir()) and path.relative_to(path.parent).match(
                        parts[0],
                    ):  # XXX ?
                        async for r in path.glob(remaining_parts):
                            yield r

    async def iterdir(self):  # pylint:disable=invalid-overridden-method
        "list a directory under this path."
        if False:
            yield self
        raise NotImplementedError


class MoatDevPath(MoatPath):
    """
    This object represents a file or directory (existing or not) on the
    target board.

    This is the implementation that connects to the REPL directly.
    To actually modify the target, :meth:`~moat.micro.files.MoatPath.connect_repl`
    must have been called.
    """

    @property
    def repl(self):
        """The connected direct REPL."""
        from .direct import DirectREPL  # noqa: PLC0415

        repl = self._repl
        if not isinstance(repl, DirectREPL):
            raise TypeError("No direct REPL connected")
        return repl

    # methods that access files

    async def stat(self, *, follow_symlinks=False) -> os.stat_result:  # noqa:ARG002
        """
        Return stat information about path on remote. The information is cached
        to speed up operations.
        """
        if self._stat_cache is None:
            st = await self.repl.evaluate(
                f"import os; print(os.stat({self.as_posix()!r}))",
                quiet=True,
            )
            self._stat_cache = os.stat_result(st)
        return self._stat_cache

    async def exists(self):
        """Return True if target exists"""
        try:
            await self.stat()
        except FileNotFoundError:
            return False
        else:
            return True

    async def is_dir(self):
        """Return True if target exists and is a directory"""
        try:
            return ((await self.stat()).st_mode & stat.S_IFDIR) != 0
        except FileNotFoundError:
            return False

    async def is_file(self):
        """Return True if target exists and is a regular file"""
        try:
            return ((await self.stat()).st_mode & stat.S_IFREG) != 0
        except FileNotFoundError:
            return False

    async def unlink(self, missing_ok=False):
        """
        :raises FileNotFoundError:

        Delete file. See also :meth:`rmdir`.
        """
        self._stat_cache = None
        try:
            await self.repl.evaluate(
                f"import os; print(os.remove({self.as_posix()!r}))",
                quiet=True,
            )
        except FileNotFoundError:
            if not missing_ok:
                raise

    async def rename(self, target):
        """
        :param target: new name
        :return: new path object
        :raises FileNotFoundError: Source is not found
        :raises FileExistsError: Target already exits

        Rename file or directory. Source and target path need to be on the same
        filesystem.
        """
        self._stat_cache = None
        await self.repl.evaluate(
            f"import os; print(os.rename({self.as_posix()!r}, {target.as_posix()!r}))",
            quiet=True,
        )
        return self.with_name(target)  # XXX, moves across dirs

    async def mkdir(self, mode=0o755, parents=False, exist_ok=False):  # noqa:ARG002
        """
        :param parents: When true, create parent directories
        :param exist_ok: No error if the directory does not exist
        :param mode: ignored.
        :raises FileNotFoundError:

        Create new directory.
        """
        try:
            return await self.repl.evaluate(
                f"import os; print(os.mkdir({self.as_posix()!r}))",
                quiet=True,
            )
        except FileExistsError:
            if exist_ok:
                pass
            else:
                raise
        except FileNotFoundError:
            if parents:
                await self.parent.mkdir(parents=True)
                await self.mkdir()
            else:
                raise

    async def rmdir(self):
        """
        :raises FileNotFoundError:

        Remove (empty) directory
        """
        await self.repl.evaluate(f"import os; print(os.rmdir({self.as_posix()!r}))", quiet=True)
        self._stat_cache = None

    async def read_as_stream(self):
        """
        :returns: async Iterator
        :rtype: Iterator of bytes

        Iterate over blocks (`bytes`) of a remote file.
        """
        # reading (lines * linesize) must not take more than 1sec and 2kB target RAM!
        n_blocks = max(1, self.repl.serial.baudrate // 5120)
        await self.repl.exec(
            f'import ubinascii; _f = open({self.as_posix()!r}, "rb"); '
            "_mem = memoryview(bytearray(512))\n"
            "def _b(blocks=8):\n"
            '  print("[")\n'
            "  for _ in range(blocks):\n"
            "    n = _f.readinto(_mem)\n"
            "    if not n: break\n"
            '    print(ubinascii.b2a_base64(_mem[:n]), ",")\n'
            '  print("]")',
            quiet=True,
        )
        while True:
            blocks = await self.repl.evaluate(f"_b({n_blocks})", quiet=True)
            if not blocks:
                break
            for block in blocks:
                yield binascii.a2b_base64(block)
        await self.repl.exec("_f.close(); del _f, _b", quiet=True)

    async def read_bytes(self) -> bytes:
        """
        :returns: file contents
        :rtype: bytes

        Return the contents of a remote file as byte string.
        """
        res = []
        async for r in self.read_as_stream():
            res.append(r)
        return b"".join(res)

    async def write_bytes(self, data, chunk=512) -> int:
        """
        :param bytes contents: Data

        Write contents (expected to be bytes) to a file on the target.
        """
        self._stat_cache = None
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f"contents must be bytes/bytearray, got {type(data)} instead")
        await self.repl.exec(
            f'from binascii import a2b_base64 as _a2b; _f = open({self.as_posix()!r}, "wb")',
            quiet=True,
        )
        # write in chunks
        with io.BytesIO(data) as local_file:
            while True:
                block = local_file.read(chunk)
                if not block:
                    break
                await self.repl.exec(
                    f"_f.write(_a2b({binascii.b2a_base64(block).rstrip()!r}))",
                    quiet=True,
                )
        await self.repl.exec("_f.close(); del _f, _a2b", quiet=True)
        return len(data)

    # read_text(), write_text()

    async def iterdir(self):
        """
        Return iterator over items in given remote path.
        """
        if not self.is_absolute():
            raise ValueError(f'only absolute paths are supported (beginning with "/"): {self!r}')
        # simple version
        # remote_paths = self._repl.evaluate(
        #     f'import os; print(os.listdir({self.as_posix()!r}))', quiet=True)
        # return [(self / p).connect_repl(self._repl) for p in remote_paths]
        # variant with pre-loading stat info
        posix_path_slash = self.as_posix()
        if not posix_path_slash.endswith("/"):
            posix_path_slash += "/"
        remote_paths_stat = await self.repl.evaluate(
            'import os; print("[")\n'
            f"for n in os.listdir({self.as_posix()!r}): "
            '    print("[", repr(n), ",", os.stat({posix_path_slash!r} + n), "],")\n'
            'print("]")',
            quiet=True,
        )
        for p, st in remote_paths_stat:
            yield self.joinpath(p)._with_stat(st)  # noqa:SLF001 pylint:disable=protected-access

    # custom extension methods

    async def sha256(self, l: int | None = None) -> bytes:  # noqa: E741
        """
        :returns: hash over file contents

        Calculate a SHA256 over the file contents and return the digest.
        """
        try:
            await self.repl.exec(
                "import hashlib; _h = hashlib.sha256(); _mem = memoryview(bytearray(512))\n"
                f'with open({self.as_posix()!r}, "rb") as _f:\n'
                "  while True:\n"
                "    _n = _f.readinto(_mem)\n"
                "    if not _n: break\n"
                "    _h.update(_mem[:_n])\n"
                "del _n, _f, _mem\n",
                quiet=True,
            )
        except ImportError:
            # fallback if no hashlib is available: download and hash here.
            try:
                _h = hashlib.sha256()
                async for block in self.read_as_stream():
                    _h.update(block)
                return _h.digest()
            except FileNotFoundError:
                return b""
        except OSError:
            hash_value = b""
        else:
            if l is None:
                hash_value = await self.repl.evaluate("print(_h.digest()); del _h", quiet=True)
            else:
                hash_value = await self.repl.evaluate(
                    f"print(_h.digest()[:{l}]); del _h",
                    quiet=True,
                )
        return hash_value


class MoatFSPath(MoatPath):
    """
    This object represents a file or directory (existing or not) on the
    target board.

    This is the implementation that connects via MoaT "f*" commands.

    To actually modify the target, :meth:`~moat.micro.files.MoatPath.connect_repl`
    must have been called.
    """

    @property
    def repl(self):
        """The connected RPC sender."""
        from moat.lib.rpc import SubMsgSender  # noqa: PLC0415

        repl = self._repl
        if not isinstance(repl, SubMsgSender):
            raise TypeError("No RPC sender connected")
        return repl

    # methods that access files

    async def _req(self, cmd, *a, **kw):
        res = await self.repl.cmd(cmd, *a, **kw)
        if res.kw:
            return res.kw
        if len(res) == 1:
            return res[0]
        return res.args

    # >>> os.stat_result((1,2,3,4,5,6,7,8,9,10))
    # os.stat_result(st_mode=1, st_ino=2, st_dev=3, st_nlink=4,
    #                st_uid=5, st_gid=6, st_size=7, st_atime=8, st_mtime=9, st_ctime=10)

    async def stat(self, *, follow_symlinks=False) -> os.stat_result:  # noqa:ARG002
        """
        Return stat information about path on remote. The information is cached
        to speed up operations.
        """
        if self._stat_cache is None:
            st = await self._req("stat", self.as_posix(), v=True)
            self._stat_cache = os.stat_result(st["d"])
        return self._stat_cache

    async def exists(self):
        """Return True if target exists"""
        try:
            await self.stat()
        except FileNotFoundError:
            return False
        else:
            return True

    async def is_dir(self):
        """Return True if target exists and is a directory"""
        try:
            return stat.S_ISDIR((await self.stat()).st_mode)
        except FileNotFoundError:
            return False

    async def is_file(self):
        """Return True if target exists and is a regular file"""
        try:
            return stat.S_ISREG((await self.stat()).st_mode)
        except FileNotFoundError:
            return False

    async def unlink(self, missing_ok=False):
        """
        :raises FileNotFoundError:

        Delete file. See also :meth:`rmdir`.
        """
        self._stat_cache = None
        try:
            await self._req("rm", self.as_posix())
        except FileNotFoundError:
            if not missing_ok:
                raise

    async def rename(self, target):
        """
        :param target: new name
        :return: new path object
        :raises FileNotFoundError: Source is not found
        :raises FileExistsError: Target already exits

        Rename file or directory. Source and target path need to be on the same
        filesystem.
        """
        self._stat_cache = None
        await self._req("mv", self.as_posix(), target.as_posix())
        return self.with_name(target)  # XXX, moves across dirs

    async def mkdir(self, mode=0o755, parents=False, exist_ok=False):  # noqa:ARG002
        """
        :param parents: When true, create parent directories
        :param exist_ok: No error if the directory does not exist
        :raises FileNotFoundError:

        Create new directory.
        """
        self._stat_cache = None
        try:
            return await self._req("mkdir", self.as_posix())
        except FileNotFoundError:
            if parents:
                await self.parent.mkdir(parents=True)
                return await self._req("mkdir", self.as_posix())
            else:
                raise
        except FileExistsError:
            if exist_ok:
                pass
            else:
                raise

    async def rmdir(self):
        """
        :raises FileNotFoundError:

        Remove (empty) directory
        """
        res = await self._req("rmdir", self.as_posix())
        self._stat_cache = None
        return res

    async def read_as_stream(self, chunk=128):
        """
        :returns: async Iterator
        :rtype: Iterator of bytes

        Iterate over blocks (`bytes`) of a remote file.
        """
        fd = await self._req("open", self.as_posix(), m="r")
        try:
            off = 0
            while True:
                d = await self._req("rd", fd, off, n=chunk)
                if not d:
                    break
                off += len(d)
                yield d
        finally:
            await self._req("cl", fd)

    async def read_bytes(self, chunk=128) -> bytes:
        """
        :returns: file contents
        :rtype: bytes

        Return the contents of a remote file as byte string.
        """
        res = []
        async for r in self.read_as_stream(chunk=chunk):
            res.append(r)
        return b"".join(res)

    async def write_bytes(self, data, chunk=128) -> int:
        """
        :param bytes contents: Data

        Write contents (expected to be bytes) to a file on the target.
        """
        self._stat_cache = None
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f"contents must be a buffer, got {type(data)} instead")
        fd = await self._req("open", self.as_posix(), m="w")
        try:
            off = 0
            while off < len(data):
                n = await self._req("wr", fd, off, d=data[off : off + chunk])
                if not n:
                    raise EOFError
                off += n

        finally:
            await self._req("cl", fd)
        return len(data)

    # read_text(), write_text()

    async def iterdir(self):
        """
        Return iterator over items in given remote path.
        """
        if not self.is_absolute():
            raise ValueError(f'only absolute paths are supported (beginning with "/"): {self!r}')
        d = await self._req("dir", self.as_posix())
        for n in d:
            yield self / n
            # TODO add stat

    async def sha256(self, l: int | None = None) -> bytes:  # noqa: E741
        """
        :returns: hash over file contents

        Calculate a SHA256 over the file contents and return the digest.
        """
        return await self._req("hash", self.as_posix(), l=l)


async def sha256(p, l: int | None = None):  # noqa: E741
    """
    Calculate a SHA256 of the file contents at @p.
    """
    try:
        return await p.sha256(l=l)
    except AttributeError:
        _h = hashlib.sha256()
        if hasattr(p, "read_as_stream"):
            async for block in p.read_as_stream():
                _h.update(block)
        else:
            async with await p.open("rb") as _f:
                _h.update(await _f.read())

        res = _h.digest()
        if l is not None:
            res = res[:l]
        return res


async def _nullcheck(p):
    """Null check function, always True"""
    if await p.is_dir():
        return p.name != "__pycache__"
    if p.suffix in (".py", ".mpy", ".state"):
        return True
    logger.info("Ignored: %s", p)
    return False


async def f_eq(src, dst):
    """check whether twofiles are the same"""
    s1 = (await src.stat()).st_size
    try:
        s2 = (await dst.stat()).st_size
    except FileNotFoundError:
        s2 = -1
    except RemoteError as err:
        if err.args[0] != "fn":
            raise
        s2 = -1
    if s1 == s2:
        h1 = await sha256(src, 12)
        h2 = await sha256(dst, 12)
        if h1 != h2:
            s2 = -1

    return s1 == s2


async def copytree(
    src: APath | ABytes,
    dst: MoatPath,
    wdst: MoatPath | APath | None = None,
    check=None,
    drop=None,
    cross=None,
):
    """
    Copy a file or directory tree from @src to @dst.
    Skip files/subtrees for which "await check(src)" is False.
    (@src is never checked.)

    Files are copied if their size or content hash differs.

    Returns the number of modified files.

    @drop is an async function with the destination file as input. If it
    returns `True` the file is deleted unconditionally, `False` (the
    default) does a standard sync-and-update, `None` ignores it.

    If @wdst is set, new files get written there; they will be deleted from @dst.
    """
    n = 0
    if wdst is None:
        wdst = dst
    assert wdst is not None
    if await src.is_file():
        if src.suffix == ".py":
            # here we replace "src" with a buffer containing the
            # corresponding mpy-cross output.

            dr = False if drop is None else await drop(dst)
            if dr:
                with suppress(FileNotFoundError):
                    await dst.unlink()
                    n += 1
                with suppress(FileNotFoundError):
                    await dst.with_suffix(".mpy").unlink()
                    n += 1
                return n

            if dr is None:
                return 0

            if wdst is not dst:
                with suppress(OSError, RemoteError):
                    await dst.unlink()
                with suppress(OSError, RemoteError):
                    await dst.with_suffix(".mpy").unlink()

            if cross:
                try:
                    p = str(src)
                    if (pi := p.find("/_embed/lib/")) > 0:
                        p = p[pi + 12 :]
                    data = await run(cross, str(src), "-s", p, "-o", "/dev/stdout", capture="raw")
                except CalledProcessError as exc:
                    print(exc.stderr.decode("utf-8"), file=sys.stderr)
                    # copy this file unmodified
                else:
                    assert isinstance(src, APath)
                    src = ABytes(src.with_suffix(".mpy"), data)
                    if isinstance(wdst, APath):
                        wdst = APath(wdst.with_suffix(".mpy"))
                    else:
                        wdst = wdst.with_suffix(".mpy")
                    with suppress(OSError, RemoteError):
                        if await f_eq(src, wdst):
                            return 0
                        await wdst.unlink()
            else:
                with suppress(OSError, RemoteError):
                    await dst.with_suffix(".mpy").unlink()

        if wdst is not dst:
            print(" Write:", dst, "       ", end="\r", file=sys.stderr)
            new = True
            with suppress(OSError, RemoteError):
                await dst.unlink()
                new = False
            await wdst.parent.mkdir(parents=True, exist_ok=True)
            await wdst.write_bytes(await src.read_bytes())
            logger.info("Copy: %s: %s", "new" if new else "update", dst)
            return 1

        if await f_eq(src, dst):
            return 0

        await wdst.parent.mkdir(parents=True, exist_ok=True)
        new = True
        with suppress(OSError, RemoteError):
            await wdst.unlink()
            new = False
        await wdst.write_bytes(await src.read_bytes())
        logger.info("Copy: %s: %s", "new" if new else "update", wdst)
        return 1

    else:
        assert isinstance(src, APath)
        if dst.name == "__pycache__":
            return 0
        if dst.name.startswith("."):
            return 0
        with suppress(OSError, RemoteError):
            s = await dst.stat()

        # logger.debug("Copy: dir %s > %s", src, dst)
        print("  Sync:", dst, "       ", end="\r", file=sys.stderr)
        sys.stderr.flush()
        async for s in src.iterdir():
            if check is not None and not await check(s):
                continue
            d = dst.joinpath(s.name)
            n += await copytree(
                s,
                d,
                check=check,
                cross=cross,
                drop=drop,
                wdst=None if wdst is dst else wdst.joinpath(s.name),
            )
        return n


async def copy_over(src, dst, cross=None, wdst: MoatPath | APath | None = None):
    """
    Transfer a file tree from @src to @dst.

    This procedure verifies that the data arrived OK.
    """
    tn = 0
    if wdst is not None and not isinstance(wdst, (MoatPath, APath)):
        wdst = APath(wdst)
    if await src.is_file():
        if await dst.is_dir():
            dst /= src.name
    while n := await copytree(src, dst, cross=cross, wdst=wdst):
        tn += n
    #       if n == 1:
    #           logger.info("One file changed. Verifying.")
    #       else:
    #           logger.info("%d files changed. Verifying.", n)
    #   logger.info("Done. No (more) differences detected.")
    return tn



"helpers"

from __future__ import annotations

from moat.lib.micro import CancelledError, Event, sleep_ms
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg

    from collections.abc import Awaitable, Callable
    from typing import ClassVar

# like get/set_part but without the attributes

TEST_MAGIC = b"r:\x0dn:\x0arn:\x0d\x0a-\x00x\x0ce\x1b" + "🖖🏼  ".encode("utf-8") + b"!"


def get_p(cur, p, add=False):
    "retrieve an item"
    for pp in p:
        try:
            cur = cur[pp]
        except KeyError:
            if not add:
                raise
            cur[pp] = nc = {}
            cur = nc
    return cur


def set_p(cur, p, v):
    "set an item"
    cur = get_p(cur, p[:-1], add=True)
    cur[p[-1]] = v


def del_p(cur, p):
    "delete an item"
    pp = p[0]
    if pp in cur:
        if len(p) > 1:
            del_p(cur[pp], p[1:])
        if cur[pp]:
            return
        del cur[pp]


class Repeater:
    """
    Support for repeating a measurement.

    Parameters:
        min(float): Minimum value.
        max(float): Maximum value.
        retry(int): #Retries before erroring. Zero=infinite.
        timer(int): Time between retries (ms)
        diff(float): Maximum difference between old and new values.

    Arguments:
        cfg: Config data.
        rdr: The async procedure to retrieve the value.
        min: default minimum
        max: default maximum
        diff: default max delta
    """

    val: float | None = None
    evt: Event | bool = False
    err: float | None = None
    exc: Exception | None = None
    rr: int = 0
    retries: int = 0

    def __init__(
        self,
        cfg: dict,
        rdr: Callable[[], Awaitable[float]],
        min: float = -99999,  # noqa:A002
        max: float = 99999,  # noqa:A002
        diff: float = 99999,
    ) -> None:
        self.cfg = cfg
        self.rdr = rdr
        self.min = min
        self.max = max
        self.diff = diff

    def state(self) -> dict:
        "Current state"
        res = dict(r=self.retries, rr=self.rr, val=self.val)
        if self.err is not None:
            res["err"] = self.err
            self.err = None
        if self.evt:
            res["wait"] = self.evt is not True
        return res

    async def get(self, force: bool = False):
        """
        Read the next bit. Ignore the *diff* parameter if *force* is
        ``True``.
        """

        # This dance ensures that an event is only allocated when more than
        # one task tries to read at the same time

        if force:
            return await self.rdr()

        if self.evt is True:
            self.evt = Event()
        if isinstance(self.evt, Event):
            while isinstance(self.evt, Event):
                await self.evt.wait()
            return self.val

        try:
            self.evt = True
            self.rr = 0
            rep = self.cfg.get("repeat", 3)
            val = None
            err = None
            while True:
                try:
                    val = await self.rdr()
                except Exception as exc:
                    if err is None:
                        err = exc
                if val is not None and self.cfg.get("min", self.min) < val < self.cfg.get(
                    "max",
                    self.max,
                ):
                    if self.val is None or abs(val - self.val) < self.cfg.get("diff", self.diff):
                        break
                self.err = val
                if rep:
                    rep -= 1
                    if rep == 0:
                        raise err or ValueError(val)
                self.rr += 1
                self.retries += 1
                await sleep_ms(self.cfg.get("timer", 20))
            if self.retries > 0:
                self.retries -= 1
        except BaseException:
            val = None
            raise
        finally:
            if isinstance(self.evt, Event):
                self.evt.set()
            self.evt = False
        if val is None:
            raise err or CancelledError
        self.val = val
        return val


class Sensor(BaseCmd):
    """
    This is the base class for a simple (one-value) sensor.

    The parameters from `Repeater` are used.

    You might want to override:
    - MIN: default minimum value
    - MAX: default maximum value
    """

    _wants: Event
    _gets: Event
    val: float
    retries: int = 0

    MIN: ClassVar[int | float] = -999999
    MAX: ClassVar[int | float] = 999999

    async def setup(self):
        "Allocate events"
        await super().setup()
        self._rep = Repeater(self.cfg, self.read, min=self.MIN, max=self.MAX)

        self._wants = Event()
        self._gets = Event()

    doc = dict(
        _c=dict(
            _d="Sensor base",
            min="float:min value",
            max="float:max value",
            retry="int:before error; zero=inf",
            timer="int:error delay(ms)",
        ),
        _d="read",
        t="int:ms between reads when streaming, default 10s",
        o="bool:old: wait until value differs",
    )

    async def read(self):
        """
        Read a value. Must be overridden.
        """
        raise NotImplementedError

    doc_st = dict(
        _d="Status", _r=dict(val="float:last value", r="int:Retries", rr="int:current retries")
    )

    async def cmd_st(self):
        "Status"
        return self._rep.state()

    async def stream(self, msg: Msg):
        "Wait for change if @o (old value) is not None"
        t = msg.get("t", 10000)
        o = msg.get("o", None)
        d = msg.get("d", 0)
        if msg.can_stream:
            async with msg.stream_out() as m:
                while True:
                    val = await self._rep.get()
                    if o is None or abs(val - o) > d:
                        await m.send(val)
                        if o is not None:
                            o = val
                    await sleep_ms(t)

        val = await self._rep.get(msg.get("force", False))
        while o is not None and abs(val - o) <= d:
            await sleep_ms(t)
            val = await self._rep.get(msg.get("force", False))
        await msg.result(val)



"""
Console stdio access
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.cmd.msg import BaseCmdMsg
from moat.lib.stream import FilenoBuf, serial_stack


class StdIO(BaseCmdMsg):
    """Sends/receives MoaT messages using stdin/stdout"""

    doc = dict(_c=dict(_d="link via stdandard i/o"))

    async def stream(self):  # noqa:D102
        cs = FilenoBuf(self.cfg, 0, 1)
        return await AC_use(self, serial_stack(cs, self.cfg))



"""
Stream link-up support for MoaT commands
"""

from __future__ import annotations

import sys

from moat.util import attrdict, merge
from moat.lib.codec.errors import SilentRemoteError
from moat.lib.micro import AC_use, BaseExceptionGroup, L, TaskGroup, idle, log  # noqa:A004
from moat.lib.rpc import BaseCmd, HandlerStream

__all__ = ["BaseCmdMsg", "CmdMsg", "ExtCmdMsg", "MsgStream", "SingleCmdMsg"]

# Typing
from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Auth, BaseMsgHandler
    from moat.lib.rpc.msg import Msg
    from moat.lib.stream import BaseMsg
    from moat.lib.stream.base import Buffer, MutBuffer

    from collections.abc import Mapping, Sequence
    from typing import Any, Protocol

    class _ConsoleMsgProto(Protocol):
        async def crd(self, buf: MutBuffer) -> int: ...

        async def cwr(self, buf: Buffer) -> None: ...


class MsgStream(HandlerStream):
    """
    This :moat.lib.rpc.stream:`HandlerStream` subclass
    interfaces with a `BaseCmdMsg` stream.

    """

    def __init__(self, handler: BaseMsgHandler, stream: BaseMsg):
        self.__stream = stream
        super().__init__(handler)

    async def read_stream(self):
        "Background stream reader. Started from the HandlerStream context manager."
        str = self.__stream  # noqa: A001
        while True:
            msg = await str.recv()
            await self.msg_in(msg)

    async def write_stream(self):
        "Background stream writer. Started from the HandlerStream context manager."
        str = self.__stream  # noqa: A001
        while True:
            msg = await self.msg_out()
            await str.send(msg)


class BaseCmdMsg(BaseCmd):
    """
    This is a command handler that relays arbitrary MoaT-RPC messages
    and a `~moat.lib.stream.BaseMsg`-compatible stream.

    The difference to `~moat.lib.rpc.stream.cmdbbm.BaseCmdBBM` is that this
    class encapsulates arbitrary message/stream calls and requires a
    `~moat.lib.rpc.cmd.msg.BaseCmdMsg` handler on the other side to talk to.

    In contrast, a `~moat.lib.rpc.stream.cmdbbm.BaseCmdBBM` exposes commands
    that directly read or write the underlying stream (of whatever type).

    This class cannot wrap a pre-existing stream. Its :meth:`stream` method
    **must** be overridden to create the stream.

    Parameters:
        prefix.recv(Path): Prefix for incoming messages
        prefix.send(Path): Prefix for outgoing messages

    If the configuration has an ``auth`` item, this will redirect
    all messages through a :py.cls:`~moat.lib.rpc.Auth` instance.
    """

    tg: object | None = None
    __stream = None
    __rprefix = ()
    stream_owner_obj_: object

    doc = dict(_d="Foo")
    auth: attrdict | None = None
    is_server: bool = False
    _auth: Auth | None = None

    auth_name: str | None = None

    def __init__(self, cfg, *, is_server: bool = False, **kw):
        self.is_server = is_server
        super().__init__(cfg, **kw)
        if "auth" in cfg:
            from moat.lib.rpc import Auth  # noqa:PLC0415

            self.auth = attrdict()
            if "pytest" in sys.modules:
                tcfg = cfg.auth.get("test", None)
                if tcfg is not None:
                    self.auth.update(tcfg)
            self._auth = Auth(cfg.auth, self)

    @property
    def auth_helper(self) -> Auth | None:
        "Getter."
        return self._auth

    def auth_stop(self) -> None:
        """
        Stop auth processing.

        This does not remove the auth handler due to security
        considerations.
        """
        if self._auth is None:
            return
        self._auth.stop()

    def auth_data_out(self) -> dict:
        """
        Generates data for the outgoing auth message.

        Called on auth startup.
        """
        return {}

    def auth_data_in(self, args: Sequence, data: Mapping[str, Any]) -> None:
        """
        Data from the incoming auth message.
        """
        args  # noqa:B018
        data  # noqa:B018
        pass

    def auth_data_res_out(self, role: str) -> dict:
        """
        Generates data for the outgoing auth acknowledgment.

        Called after auth is complete.

        Args:
            role: the auth method that succeeded.
        """
        role  # noqa:B018
        return {}

    def auth_data_res_in(self, role: str, data: Mapping[str, Any]) -> None:
        """
        Data from the incoming auth acknowledgment.

        Args:
            role: the auth method that succeeded on the remote side.
            data: other data that the remote side returned.
        """
        pass

    def auth_skip(self) -> None:
        """
        Callback to signal that auth did not take place because the other
        side doesn't support it.
        """
        pass

    async def wait_for_auth(self):
        "wait for auth to complete"
        if self._auth is not None:
            await self._auth.wait_done()

    def stream_owner_(self):
        """Owner for stream contexts opened by :meth:`stream`.

        Auth temporarily overrides this so transport-layer contexts opened in
        ``process()`` are closed before the auth taskgroup exits.
        """
        return getattr(self, "stream_owner_obj_", self)

    async def teardown(self):
        "also cancel auth"
        self.auth_stop()
        await super().teardown()

    async def stream(self) -> BaseMsg:
        """
        This method creates (and returns) the data stream.

        Must be overridden.

        Cleanup is typically handled via `moat.lib.micro.AC_use`.
        """
        raise NotImplementedError("Create the stream: ", self.__class__.__name__)

    async def setup(self):
        "Sets ``__rprefix``"
        await super().setup()

        rprefix = self.cfg.get("prefix", {}).get("send", ())
        if rprefix:
            rprefix = list(rprefix)
            rprefix.reverse()
            self.__rprefix = rprefix

    async def task(self):
        """
        Start the MsgStream.
        """
        root0 = self.root
        if root0 is None:
            raise RuntimeError("Not attached")
        root = root0.sender
        lprefix = self.cfg.get("prefix", {}).get("recv", ())
        if lprefix:
            root = root.sub_at(lprefix)

        if self._auth:
            await self._auth.process(root)
        else:
            await self.process(root)

    async def process(self, root: BaseMsgHandler):
        """
        Low-level handler to run the message processor.

        Args:
            root: The MsgSender to route incoming commands to.

        The stream to use is retrieved by calling :py.meth:`stream`.
        """
        try:
            self.s = await self.stream()
            async with MsgStream(root, self.s) as st:
                self.__stream = st
                if L:
                    self.set_ready()
                await idle()

        finally:
            self.s = None
            self.__stream = None

    async def handle(
        self, msg: Msg, rcmd: list[PathElem], *prefix: str, _auth: bool = False
    ) -> Any:
        """
        Forward a request to some remote side.
        """
        prefix  # noqa:B018
        # If auth, route through it.
        if self._auth and not _auth:
            return await self._auth.handle(msg, rcmd)

        # Handle local commands (and documentation) locally
        if (
            (len(rcmd) == 1 or (len(rcmd) == 2 and rcmd[1] == "doc_"))
            and rcmd[0] != "dir_"
            and (hasattr(self, f"cmd_{rcmd[0]}") or hasattr(self, f"stream_{rcmd[0]}"))
        ):
            return await super().handle(msg, rcmd)

        await self.check_rdy(msg, rcmd)
        if self.__stream is None:
            raise EOFError

        # forward to remote
        rcmd.extend(self.__rprefix)
        res = await self.__stream.handle(msg, rcmd)

        # if it was a directory request, add local data
        if len(rcmd) == 1 and rcmd[0] == "dir_":
            # Merge local commands into the remote ones
            m2 = await self.cmd_dir_(v=msg.get("v", True))
            await msg.wait_replied(preload=True)

            if msg._kw is None:  # noqa: SLF001
                msg._kw = {}  # noqa: SLF001
            kw = cast(dict, msg._kw)  # noqa: SLF001
            kw["c"] = tuple(set(msg.get("c", ())) | set(m2.pop("c", ())))
            kw["s"] = tuple(set(msg.get("s", ())) | set(m2.pop("s", ())))
            merge(kw, m2)
        return res

    doc_crd = dict(_d="read console", _0="int:len (64)")

    async def cmd_crd(self, n=64) -> Buffer:
        """read some console data"""
        b = bytearray(n)
        if self.s is None:
            raise EOFError
        s = cast("_ConsoleMsgProto", self.s)
        r = await s.crd(b)
        if r == n:
            return b
        elif r <= n >> 2:
            return bytes(b[:r])
        else:
            b = memoryview(b)
            return b[:r]

    doc_cwr = dict(_d="write console", _0="bytes:data")

    async def cmd_cwr(self, b: Buffer):
        """write some console data"""
        if self.s is None:
            raise EOFError
        await cast("_ConsoleMsgProto", self.s).cwr(b)

    doc_c = dict(
        _d="r/w console stream", _0="int:rdbuflen (64)", _i="bytes:to send", _o="bytes:received"
    )

    async def stream_c(self, msg):
        "read/write console stream"
        n = msg.get(0, 64)
        async with msg.stream() as st, TaskGroup() as tg:

            @tg.start_soon
            async def crd():
                while True:
                    await st.send(await self.cmd_crd(n))

            async for m in st:
                await self.cmd_cwr(m[0])
            tg.cancel()


class CmdMsg(BaseCmdMsg):
    """
    A baseCmdMsg with a ready-made link that it opens.
    """

    def __init__(self, cfg: dict, link: BaseMsg):
        super().__init__(cfg)
        self.link = link

    async def stream(self) -> BaseMsg:  # noqa:D102
        # pylint:disable=invalid-overridden-method
        return await AC_use(self.stream_owner_(), self.link)


class SingleCmdMsg(BaseCmdMsg):
    """
    A BaseCmdMsg that disconnects on error, or when the connection ends,
    without propagating the exception.
    """

    # pylint:disable=abstract-method
    # `stream` needs to be implemented by a subclass

    async def run(self):  # noqa:D102
        # this would be far easier with "except*"
        # but µPy doesn't have that.
        try:
            try:
                await super().run()
            except BaseExceptionGroup as e:
                while True:
                    if len(e.exceptions) != 1:
                        a, b = e.split((EOFError, OSError, SilentRemoteError))
                        if a is not None:
                            log("Err %s: %r", self.path, repr(a))
                        if b is None:
                            return
                        raise b  # noqa:B904,RUF100
                    e = e.exceptions[0]
                    if not isinstance(e, BaseExceptionGroup):
                        raise e  # noqa:TRY201
        except EOFError:
            pass
        except (OSError, SilentRemoteError) as exc:
            log("Err %s: %r", self.path, repr(exc))
        except Exception as exc:  # pylint:disable=broad-exception-caught
            log("Err %s", self.path, err=exc)


class ExtCmdMsg(SingleCmdMsg):
    """SingleCmdMsg, on a stream that was established externally.

    The caller is responsible for calling :meth:`~moat.lib.rpc.BaseCmd.wait_stopped`
    and then closing the stream!
    """

    def __init__(self, cfg: dict[str, Any], stream: BaseMsg, *, is_server: bool = False):
        if cfg is None:
            cfg = {}
        super().__init__(cfg, is_server=is_server)
        self.__s = stream

    async def stream(self):  # noqa:D102
        return await AC_use(self.stream_owner_(), self.__s)



"""
Command tree support for MoaT commands
"""

from __future__ import annotations

from moat.util import attrdict
from moat.lib.micro import L, TaskGroup
from moat.lib.rpc import LoadCmd

from .dir import BaseSuperCmd

# Typing
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import BaseCmd, Msg


class BaseLayerCmd(BaseSuperCmd):
    """
    A handler for a single nested app.

    This handler doesn't affect the command hierarchy.
    Its own commands, if any, are reachable by prepending "!" to the command.

    You need to override "gen_cmd" to create the app object.
    """

    app: BaseCmd | None = None
    name: str = "_"

    async def run_app(self):
        """
        The command handler's executor. By default, calls ``self.app.run``
        within the command's context.

        You might override this e.g. for restarting or
        shielding the rest of MoaT from errors.
        """
        if self.app is None:
            raise RuntimeError("No app")
        await self.app.run()

    async def task(self):
        """
        Run the app as a subtask.

        You typically don't override this.
        """
        async with TaskGroup() as tg:
            app = self.app
            if app is not None:
                tg.start_soon(self.run_app)
            if L:
                if app is None:
                    raise RuntimeError("No app")
                await app.wait_ready()
                self.set_ready()

            # await self.app.stopped()
            # the return from the taskgroup already does that

    async def setup(self):
        "attach the nested app"
        await super().setup()
        self.app = await self.gen_cmd()
        if self.app is not None:
            self.app.attached(self, self.name)
            if L:
                self.set_ready()

    async def reload(self):
        "reload the nested app"
        await super().reload()
        if self.app is not None:
            await self.app.reload()

    if L:

        async def wait_ready(self, wait=True):
            "wait for nested app"
            if await super().wait_ready(wait=wait):
                return True
            if self.app is None:
                return None
            return await self.app.wait_ready(wait=wait)

    async def gen_cmd(self) -> BaseCmd | None:
        """
        Create the actual app to use.

        The default uses `None` and leaves the setup to `task`.
        """
        return None

    async def handle(self, msg: Msg, rcmd: list[PathElem]):
        """
        Forward to the sub-app.

        The subcommand "!" redirects to the local handler.
        """
        if rcmd and isinstance(rcmd[-1], str) and rcmd[-1] == "!":
            rcmd.pop()
            return await super().handle(msg, rcmd)

        if L:
            await self.wait_ready()
        if self.app is None:
            raise RuntimeError("No app")
        return await self.app.handle(msg, rcmd)

    if L:

        def set_ready(self):
            "guards that the nested app exists"
            if self.app is None:
                raise RuntimeError("early")
            super().set_ready()

    def __getattr__(self, k):
        if k.startswith("_"):
            raise AttributeError(k)
        if k.endswith("_f"):
            return getattr(self, k[:-2])
        return getattr(self.app, k)


class BaseFwdCmd(BaseLayerCmd):
    """
    A handler for a single nested app that's configured locally.
    """

    async def gen_cmd(self):
        """
        Create the underlying app object
        """
        gcfg = self.cfg
        cfg = gcfg.get("cfg", attrdict())

        # log("Setup %s: %s", self.path, cfg["app"])
        return LoadCmd(cfg)



"""
Exception handling helpers

This code is *duplicated* in `moat.util.exc`.
"""

from __future__ import annotations

from moat.lib.micro import BaseExceptionGroup  # noqa:A004

__all__ = [
    "ExpAttrError",
    "ExpKeyError",
    "ExpectedError",
    "exc_iter",
    "ungroup",
]


class ExpKeyError(KeyError):
    "unreported key error"

    pass


class ExpAttrError(AttributeError):
    "unreported key error"

    pass


ExpectedError = (ExpKeyError, ExpAttrError)


def exc_iter(exc):
    """
    iterate over all non-exceptiongroup parts of an exception(group)
    """
    if isinstance(exc, BaseExceptionGroup):
        for e in exc.exceptions:
            yield from exc_iter(e)
    else:
        yield exc


class ungroup:
    """
    A sync+async context manager that unwraps single-element
    exception groups.
    """

    def __call__(self):
        "Singleton. Returns itself."
        return self

    @staticmethod
    def one(e):
        "convert the exceptiongroup @e to a single exception"
        while isinstance(e, BaseExceptionGroup):
            if len(e.exceptions) != 1:
                break
            e = e.exceptions[0]
        return e

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, c, e, t):
        if e is None:
            return
        e = self.one(e)
        raise e from None

    async def __aexit__(self, c, e, t):
        return self.__exit__(c, e, t)


ungroup = ungroup()



"""
Console stdio access
"""

from __future__ import annotations

import sys
from functools import partial

import micropython

from moat.lib.micro import AC_use
from moat.lib.rpc.cmd.msg import BaseCmdMsg
from moat.lib.rpc.conn.file import FileBuf
from moat.lib.stream import serial_stack

from ._doc import _link_d, _log_d

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.micro.console import Cmd


class StdioBuf(FileBuf):
    "direct access to stdio"

    doc = dict(_c=dict(_d="stdio data access"))

    async def stream(self):
        "Create a dedicated stdin/stdout stream"
        stdin = stdout = None
        try:
            stdin = open("/dev/stdin", "rb")  # noqa:SIM115,ASYNC230
            stdout = open("/dev/stdout", "wb")  # noqa:SIM115,ASYNC230
        except OSError:
            if stdin is not None:
                stdin.close()
            if stdout is not None:
                stdout.close()
            return sys.stdin.buffer, sys.stdout.buffer
        else:
            AC_use(self, stdin.close)
            AC_use(self, stdout.close)
            return stdin, stdout


class StdIO(BaseCmdMsg):
    """Sends/receives MoaT messages using stdin/stdout"""

    doc = dict(_c=dict(_d="stdio RPC access", link=_link_d, **_log_d))

    async def stream(self):
        "Set up a MoaT message stream on stdin+stdout"
        cs = StdioBuf(self.cfg)
        micropython.kbd_intr(-1)
        await AC_use(self, partial(micropython.kbd_intr, 3))
        return await AC_use(self, serial_stack(cs, self.cfg))


def console(*a, **kw) -> Cmd:
    """Creates a 'real' Python console running the REPL"""
    from moat.micro.console import Cmd  # noqa:PLC0415

    return Cmd(*a, **kw)



"""
CBOR message handling for streams (CPython-specific).
"""

from __future__ import annotations

from moat.lib.codec import get_codec

from ._cbor import _CBORMsgBlk, _CBORMsgBuf


class CBORMsgBuf(_CBORMsgBuf):
    """
    structured messages > bytestream

    Use this if your stream is reliable (TCP, USB, …) but doesn't support
    message boundaries.
    """

    async def setup(self):  # noqa:D102
        await super().setup()
        self.codec = get_codec(self.cfg.get("codec", "std-cbor"))


class CBORMsgBlk(_CBORMsgBlk):
    """
    structured messages > chunked bytestrings

    Use this if the layer below supports byte boundaries
    (one bytestring-ized message per call).
    """

    async def setup(self):  # noqa:D102
        await super().setup()
        self.codec = get_codec(self.cfg.get("codec", "std-cbor"))



"""
A command that accesses a row of mostly-identical subcommands
"""

from __future__ import annotations

from moat.util import attrdict, combine_dict, set_part
from moat.lib.codec.errors import NoPathError
from moat.lib.micro import L, TaskGroup
from moat.lib.path import Path
from moat.lib.rpc import MsgSender, ShortCommandError, load_app

from .tree.dir import BaseSuperCmd

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg

    from collections.abc import Awaitable


class ArrayCmd(BaseSuperCmd):
    """
    A command that hosts a number of mostly-identical subcommands.
    """

    n: int | None = None

    doc = dict(
        _c=dict(
            _d="subdirectory", cfg="cfg:sub-app", n="int:nr.apps", _n="dict:cfg update for app #i"
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        self.apps = []

    async def setup(self):  # noqa:D102
        await super().setup()
        await self._setup_apps()

    if L:

        async def wait_ready(self, wait=True):
            """Delay until this subtree is up,

            Returns True if any sub-apps are stopped.
            """
            await super().wait_ready(wait=wait)
            again = True
            res = False
            while again:
                again = False
                for app in self.apps:
                    if (w := await app.wait_ready(wait=wait)) is None:
                        if not wait:
                            return None
                        again = True
                        res = None
                    elif w is True:
                        return True
            return res

    async def attach(self, name, app) -> None:
        """
        Attach a sub-handler to me.

        An existing handler with this name is stopped.
        """
        try:
            oa = self.apps[name]
        except IndexError:
            if len(self.apps) != name:
                raise
            self.apps.append(app)
            oa = None
        else:
            self.apps[name] = app
        app.attached(self, name)
        if oa is not None:
            await oa.stop()

    def detach(self, name) -> Awaitable:
        """
        Detach and stop a command handler.
        """
        return self.attach(name, None)

    def _cfg(self, i):
        cfg = combine_dict(self.cfg.get("cfg", {}), self.cfg.get(i, {}), cls=attrdict)
        if (ii := self.cfg.get("i", None)) is not None:
            set_part(cfg, ii, i + self.cfg.get("i_off", 0))
        return cfg

    async def reload(self):  # noqa:D102
        await super().reload()
        self.n = self.cfg["n"]
        for i, app in enumerate(self.apps):
            app.cfg.merge(self._cfg(i))
            await app.reload()
        while len(self.apps) > self.n:
            app = self.apps.pop()
            await self.detach(len(self.apps))

    async def _setup_apps(self):
        name = self.cfg["cfg"]["app"]
        cls = load_app(name)

        self.n = self.cfg["n"]
        for i in range(self.n):
            app = cls(self._cfg(i))
            await self.attach(i, app)

        for app in self.apps:
            await self.start_app(app)

    async def handle(self, msg: Msg, rcmd: list[PathElem]):
        """
        Dispatch a message.

        Subcommands:

        * ``!`` handles the command directly.
        * ``*`` addresses all subcommands and returns an array/stream
                of the results.
        * ``n`` (integer) addresses the *n*th subcommand.

        See :meth:`moat.lib.rpc.MsgHandler.handle` for further details.
        """
        if not rcmd:
            raise ShortCommandError(msg.cmd)
        if isinstance(rcmd[-1], str) and rcmd[-1] == "!":
            rcmd.pop()
            return await super().handle(msg, rcmd)

        cmd = rcmd.pop()
        if cmd == "all":
            if msg.can_stream:
                return await self._stream_all(msg, rcmd)
            else:
                return await self._cmd_all(msg, rcmd)

        if not isinstance(cmd, int):
            raise NoPathError(
                self.path,
                msg.cmd,
                self.__class__.__name__,
                await self.cmd_dir_(v=None),
            ) from None

        try:
            sub = self.apps[cmd]
        except (TypeError, IndexError):
            raise NoPathError(
                self.path,
                msg.cmd,
                self.__class__.__name__,
                await self.cmd_dir_(v=None),
            ) from None
        return await sub.handle(msg, rcmd)

    doc_dir_ = dict(na="int:max index")

    async def cmd_dir_(self, v=True):
        "report max index"
        res = await super().cmd_dir_(v=v)
        res["na"] = len(self.apps)
        return res

    doc_all = dict(
        _d="apply to all",
        _0="path:command",
        _99="list:args",
        s="int:start index",
        e="int:end index",
    )

    async def _cmd_all(self, msg, rcmd):
        """
        Call all sub-apps and collect the result.
        """
        if rcmd:
            cmd = rcmd[:]
            cmd.reverse()
        else:
            cmd = msg.args_l.pop(0)
            if isinstance(cmd, str):
                cmd = [cmd]
            else:
                cmd = list(cmd)

        res = []
        snd = MsgSender(self)
        for app in self.apps:
            snd.set_root(app)
            r = await snd.cmd(Path.build(cmd), *msg.args, **msg.kw)
            res.append((r.args, r.kw))
        await msg.result(*res)

    async def _stream_all(self, msg, rcmd):
        """
        Call all sub-apps and send the results.
        """
        from moat.lib.rpc import Msg  # noqa:PLC0415

        if not rcmd:
            cmd = msg.args.pop(0)
            if isinstance(cmd, str):
                cmd = [cmd]
            else:
                cmd = list(cmd)
            cmd.reverse()
            rcmd = cmd

        async def _reply(i, app, st):
            msg_ = Msg.Call(msg.cmd, msg.args, msg.kw)
            res = await app.handle(msg_, rcmd[:])
            await st.send(i, *res.args, **res.kw)

        async with msg.stream_out() as st, TaskGroup() as tg:
            for i, app in enumerate(self.apps):

                async def _run(i=i, app=app) -> None:
                    await _reply(i, app, st)

                tg.start_soon(_run)

        await msg.result()



"""
Asyncio stream adaptors for MoaT (MicroPython-specific).
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.stream import BaseBuf


class AIOBuf(BaseBuf):
    """
    Adapts an asyncio stream to MoaT.

    Implement an async context handler @stream to set the stream up
    (and close it when done).
    """

    s = None

    def __init__(self):
        pass

    async def stream(self):  # noqa:D102
        raise NotImplementedError

    async def wr(self, buf):
        "translates to ``.write`` + ``.drain``"
        self.s.write(buf)
        await self.s.drain()
        return len(buf)

    async def rd(self, buf):
        "translates to ``.readinto``"
        s = self.s
        res = await s.readinto(buf)
        if not res:
            raise EOFError
        return res


class SingleAIOBuf(AIOBuf):
    """
    Adapts an asyncio stream to MoaT.

    The stream is passed to the class constructor and can only be used
    once.
    """

    def __init__(self, stream):
        self._s = stream

    async def stream(self):  # noqa:D102
        if self._s is None:
            raise RuntimeError("used twice")
        s, self._s = self._s, None
        await AC_use(self, self._destr)
        return s

    def _destr(self):
        if hasattr(self.s, "deinit"):
            self.s.deinit()
        elif hasattr(self.s, "close"):
            self.s.close()
